package adapters;

import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.Properties;

import basic.Status;

/**
 * Abstract class for interfacing with CEP engines. 
 * Vendor-specific adapters must inherit from this class. 
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public abstract class CEPEngineInterface {
	protected Status status;
	protected Properties connProperties;	
	
	/**
	 * Connects to CEP engine
	 * 
	 * @return
	 * @throws Exception
	 */
	public abstract boolean connect() throws Exception;
	
	
	/**
	 * 
	 * Performs any vendor-specific initialization at client side (e.g. initialize listeners for output streams)
	 * 
	 * @param outputToMonitor	Maps output streams at CEP engine to corresponding address and port of a Sink
	 * @return
	 * @throws Exception
	 */
	public abstract boolean load(HashMap<String, InetSocketAddress> outputToMonitor) throws Exception;
	
	
	/**
	 * Disconnects from CEP engine and performs any finalization procedures needed (e.g. close listeners for output streams)
	 * 
	 */
	public abstract void disconnect();
	
	
	/**
	 * 
	 * Sends an event to an input stream at CEP engine
	 * Converts the framework CSV-based representation to a format supported by the CEP engine
	 * 
	 * @param e
	 * @throws Exception
	 */
	public abstract void sendEvent(String e) throws Exception;
	
	
	/**
	 * Optional function. Retrieves the list of input streams of a continuous query running on CEP engine
	 * 
	 * @return
	 * @throws Exception
	 */
	public abstract String[] getInputStreamList() throws Exception;
	
	
	/**
	 * Optional function. Retrieves the list of output streams of a continuous query running on CEP engine
	 * 
	 * @return
	 * @throws Exception
	 */
	public abstract String[] getOutputStreamList() throws Exception;
}
