package adapters;

import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.Socket;
import basic.Globals;

import com.coral8.toolbox.Toolbox;
import com.coral8.tuple.Message;
import com.coral8.tuple.MessageBundle;
import com.coral8.tuple.Tuple;
import com.coral8.tuple.stream.Subscription;

/**
 * Thread for listening incoming events from an output stream at Coral8 CEP engine. 
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public class Coral8EventListener extends Thread{

	private String streamURL;
	private String streamName;
	private Subscription sub;
	
	private String sinkAddress;
	private int sinkPort;
	private Socket sinkSocket;
	ObjectOutputStream sinkOutput;
	private boolean keepListening = true;
	
	public long receivedEventCount = 0; // Test
	
	/**
	 * 
	 * @param streamURL			URL of output stream
	 * @param sinkAddress		Address of the Sink to which events from that stream must be forwarded
	 * @param sinkPort			Port of sink
	 */
	public Coral8EventListener(String streamURL, String sinkAddress, int sinkPort) {
		this.streamURL = streamURL;
		this.streamName = streamURL.substring(streamURL.lastIndexOf("/")+1);
		this.sinkAddress = sinkAddress;
		this.sinkPort = sinkPort;		
	}
	
	
	/**
	 * Stops listening for events
	 */
	public void close(){
		this.keepListening = false;		
	}
	
	/**
	 * Unsubscribes from output stream
	 * Closes the communication with the Sink
	 * 
	 * @return
	 */
	private boolean disconnect() {
		boolean ret = false;
		if(this.sub != null && !sub.hasMessagesPending()) 			
		{
			sub.disconnect();
			ret = true;
		}

		try {
			if(this.sinkOutput != null) {
				sinkOutput.close();
			}
			
			if(this.sinkSocket != null) {
				sinkSocket.close();
			}
			
		} catch (IOException e) {
			System.err.println("Could not close communication with sink.");
		}
					
		return ret;
	}
	
	/**
	 * Subscribes to an output stream on Coral8
	 * 
	 * @throws Exception
	 */
	public void load() throws Exception {						
		try {
			sub = Toolbox.subscribeToStream(this.streamURL);	
		} catch (Exception e) {
			throw new Exception("Could not subscribe to stream " + streamURL + "\n(" + e.getMessage()+")");
		}
		
	}
	
	@Override
	public void run() {					
		Message m = null;

		// Receive incoming events 
		try {
			while (!Thread.currentThread().isInterrupted() && keepListening) {
				try {
					m = sub.getNextMessage();				
					this.processIncomingEvent(m);
				} catch (InterruptedException e) {
					System.err.println("Error while listening to events. (" + e.getMessage() + ")");					
				}             
			}
		}
		finally {
			disconnect();
		}

	}
	
	/**
	 * Process events from the Output streams:
	 * Translates the event from the CEP engine native format to the framework's CSV format and forwards to Sink
	 * 
	 * @param m		The incoming message
	 */
	public void processIncomingEvent(Message m) {
		// SINGLE EVENT
		if (m instanceof Tuple) {
			Tuple t = (Tuple)m;
			
			String e = "type:" + streamName;
			for (int i = 0; i < t.getTupleDescriptor().getFieldCount(); i++) {
				e+=Globals.CSV_SEPARATOR+t.getValueAt(i);
			}			
			
			// Forward event to sink
			forward(e);
		}
		// BUNDLE OF EVENTS
		else if (m instanceof MessageBundle){			
			MessageBundle bundle = (MessageBundle) m;
			for (int i = 0; i < bundle.getMessageCount(); i++) {
				this.processIncomingEvent(bundle.getMessageAtPosition(i));
			}
		}
	}
	
	private void forward(String e) {
		// If it is the first event, open connection to sink		
		if(sinkSocket == null || sinkOutput == null || sinkSocket.isClosed()) {			
			try {
				connectTosink();
			} catch (IOException ioe1) {
				System.err.println("Could not connect to sink (" + ioe1.getMessage() + "). Trying reconnection...");				
				forward(e);
			}			
		}
		
		try {
			sinkOutput.writeObject(e);
			this.receivedEventCount++;
			if(this.receivedEventCount%100000==0)
				sinkOutput.reset();
		} catch (IOException ioe) {
			System.err.println("Error while forwarding event to sink. ("+ ioe.getMessage() + "). Trying reconnection...");			
			try {
				connectTosink();
				forward(e);
			} catch (IOException ioe1) {
				System.err.println("sink seems to be offline.");
				disconnectFromsink();
			}	
		}

	}	
	
	private void connectTosink() throws IOException {
		disconnectFromsink();
		this.sinkSocket = new Socket(sinkAddress, sinkPort);		
		this.sinkOutput = new ObjectOutputStream(sinkSocket.getOutputStream());	
	}
	
	private void disconnectFromsink() {
		if(this.sinkOutput != null) {
			try {
				sinkOutput.close();
			} catch (IOException e) {
				System.err.println("Error while closing connection with sink (" + e.getMessage() + ")");
			}
			sinkOutput = null;
		}
		
		if(this.sinkSocket != null) {
			try {
				sinkSocket.close();
			} catch (IOException e) {
				System.err.println("Error while closing connection with sink (" + e.getMessage() + ")");
			}
			sinkSocket = null;
		}
	}
}
