package adapters;


import java.io.File;
import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.Properties;
import java.util.Map.Entry;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import basic.Globals;
import basic.Status;
import basic.Step;

import com.coral8.toolbox.Toolbox;
import com.coral8.tuple.Timestamp;
import com.coral8.tuple.Tuple;
import com.coral8.tuple.TupleDescriptor;

/**
 * Implementation of an adapter for Coral8
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public class Coral8Interface extends CEPEngineInterface {
	private String serverAddress;
	private int serverPort;
	private String workspaceName;
	private String queryModuleName;
	
	private String inputStreamsList[];
	private String outputStreamsList[];
	
	private HashMap<String, TupleDescriptor> inputStreamsSchema;
	// List of listeners for incoming events
	private Coral8EventListener[] outputListeners;
	private String prefixOutputURL;
		
	public Coral8Interface(Properties connectionProperties) {
		this.connProperties = connectionProperties;
		this.status = new Status(Step.STOPPED, 0);
	}
	
	@Override
	public String[] getInputStreamList() throws Exception {
		return this.inputStreamsList;
	}

	@Override
	public String[] getOutputStreamList() throws Exception {
		return this.outputStreamsList;
	}
	
	@Override
	/**
	 * 1) Compiles and Starts query module at Coral8 server
	 * 2) Retrieves streams list 
	 */
	public boolean connect() throws Exception{
		try {
			this.serverAddress = this.connProperties.getProperty("Server_address");
			this.serverPort = Integer.parseInt(this.connProperties.getProperty("Server_port"));
			this.workspaceName = this.connProperties.getProperty("Workspace");
			this.queryModuleName = this.connProperties.getProperty("Query_module");
			prefixOutputURL = "ccl://" + this.serverAddress + ":" + this.serverPort + "/Stream/" + this.workspaceName + "/" + this.queryModuleName + "/";
			String compilerPath =  this.connProperties.getProperty("Compiler_path");
			System.setProperty("c8.compilerPath", compilerPath);			
			
			// Retrieves schemas for streams from query module CCL file
			String queryModulePath = this.connProperties.getProperty("Query_module_path");
			
			// Compile and start the module.			
			System.out.println("Compiling Query Module...");
			Toolbox.compile(queryModulePath+queryModuleName+".ccl", queryModulePath+queryModuleName + ".ccx", workspaceName, System.out, System.err);
	        System.out.println("Done! \nStarting Query Module...");
	        Toolbox.startProgramIfNotStarted("http://"+serverAddress+":"+serverPort, workspaceName, new File(queryModulePath+queryModuleName + ".ccx"));
	        System.out.println("Done!");
			
			DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
			DocumentBuilder builder;
			builder = factory.newDocumentBuilder();
			Document doc = builder.parse(new File(queryModulePath+queryModuleName+".ccl"));			
			Element xmlFileRoot = doc.getDocumentElement();			
			Element interfaceElem = (Element)xmlFileRoot.getElementsByTagName("Interface").item(0);			
			// Input
			Element inputs = (Element)interfaceElem.getElementsByTagName("Inputs").item(0);						
			NodeList inputList = inputs.getElementsByTagName("Input");
			this.inputStreamsList = new String[inputList.getLength()];
			TupleDescriptor inputDescriptor=null;
			Element input;
			this.inputStreamsSchema = new HashMap<String, TupleDescriptor>(this.inputStreamsList.length);
			for (int i = 0; i < this.inputStreamsList.length; i++) {
				input = (Element)inputList.item(i);
				inputStreamsList[i] = input.getAttribute("Name");
				inputDescriptor = Toolbox.compileSchema(queryModulePath+input.getAttribute("StreamSchema"));
				this.inputStreamsSchema.put(inputStreamsList[i], inputDescriptor);
			}			
			// Output
			Element outputs = (Element)interfaceElem.getElementsByTagName("Outputs").item(0);						
			NodeList outputList = outputs.getElementsByTagName("Output");
			this.outputStreamsList = new String[outputList.getLength()];
			Element output;
			for (int i = 0; i < this.outputStreamsList.length; i++) {
				output = (Element)outputList.item(i);
				outputStreamsList[i] = output.getAttribute("Name");
			}
			
			this.status.setStep(Step.CONNECTED);
			return true;	
		}
		catch (NumberFormatException e) {
			throw new Exception("Invalid value on property file.");		
		}
		
	}

	@Override
	/**
	 * Subscribes to output streams on Coral8 server
	 * 	 */
	public boolean load(HashMap<String, InetSocketAddress> outputToSink) throws Exception {
		// This interface instance has already been loaded
		if(this.status.getStep() == Step.READY) {
			return true;
		}
		else { // If it is not connected yet, try to connect
			if(this.status.getStep() != Step.CONNECTED) {
				this.connect();
			}
		}	
				
		if(this.status.getStep() == Step.CONNECTED) {
			this.status.setStep(Step.LOADING);

			this.outputListeners = new Coral8EventListener[outputToSink.size()];
			String outputURL;
			int i = 0;
			for (Entry<String, InetSocketAddress> e : outputToSink.entrySet()) {
				try {
					outputURL = "ccl://" + this.serverAddress + ":" + this.serverPort + "/Stream/" +
					this.workspaceName + "/" + this.queryModuleName + "/" +  e.getKey();
					this.outputListeners[i] = new Coral8EventListener(outputURL, e.getValue().getAddress().getHostAddress(), e.getValue().getPort());
					outputListeners[i].load();

				}
				catch (Exception exc) {
					this.status.setStep(Step.ERROR);
					throw new Exception("Could not load output listener for Sink at " + 
							e.getValue().getAddress().getHostAddress() + ":" +
							e.getValue().getPort() +
							"\n(" + exc.getMessage() + ")");				
				}
				i++;
			}				
			
			this.startAllEventListeners();
			
			this.status.setStep(Step.READY);
			return true;
		}
		else {
			return false;
		}
	}
	
	/**
	 * 
	 * Makes all output stream listeners ready for receiving incoming events
	 * 
	 */
	private void startAllEventListeners(){
		for (int i = 0; i < this.outputListeners.length; i++) {
			outputListeners[i].start();
		}
	}
	
	/**
	 * 
	 * Stops query module and disconnects output listeners
	 */
	public void disconnect() {
		long receivedEvents = 0;
		for (int i = 0; i < this.outputListeners.length; i++) {				
			outputListeners[i].close();			
			receivedEvents += this.outputListeners[i].receivedEventCount;
		}
		
		System.out.println("Received Events: " +receivedEvents);
		String managerURL = "http://" + this.serverAddress + ":" + this.serverPort + "/Manager";
		try {
			Toolbox.stopProgram(managerURL, this.workspaceName, this.queryModuleName);
		} catch (Exception e1) {			
			System.err.println("WARNING: Could not disconnect from Coral8 Server.");
			e1.printStackTrace();
		}

	}

	
	public void sendEvent(String e) throws Exception {		
		String ev[] = e.split(Globals.CSV_SEPARATOR);	
		String type = ev[0].substring(5);
		TupleDescriptor td = this.inputStreamsSchema.get(type);
		if(td != null){
			Tuple t = td.makeNewTuple(Timestamp.currentTimestamp());			
			for (int i = 1; i < ev.length; i++) {				
				t.setValueAt(i-1, ev[i]);
			}				
			//System.out.println(prefixOutputURL);
			Toolbox.publishRow(this.prefixOutputURL + type, t);			
		}
		else {
			System.err.println("Unknown event type. It is not possible to send event.");
		}

	}		
}
