package adapters;

import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.Properties;
import java.util.Map.Entry;

import basic.Globals;
import basic.Status;
import basic.Step;

import com.streambase.sb.EntityType;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.StreamProperties;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;
import com.streambase.sb.client.StreamBaseClient;

/**
 * Implementation of an adapter for StreamBase CEP Engine
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public class StreamBaseInterface extends CEPEngineInterface {

	private String sbAppURI;
	private StreamBaseClient client;
	
	// Stores translations from internal format to external format 
	private HashMap<String, StreamProperties> inputStreamsSchema;	
	// Threads to listen for incoming events
	private StreamBaseListener outputListeners[];	
	
	public StreamBaseInterface(Properties connectionProperties) {
		this.status = new Status(Step.DISCONNECTED, 0);
		this.connProperties = connectionProperties;
	}
	
	
	public boolean connect() throws Exception {
		String serverAddress = this.connProperties.getProperty("Server_address");
		int serverPort = Integer.parseInt(this.connProperties.getProperty("Server_port"));
		this.sbAppURI = "sb://" + serverAddress + ":" + serverPort;
		this.client = new StreamBaseClient(sbAppURI);		
		this.status.setStep(Step.CONNECTED);
		return true;
	}
	
	protected void loadInput() throws Exception {		
		// Input
		try {				
			String inputStreamsNames[] = this.getInputStreamList();
			inputStreamsSchema = new HashMap<String, StreamProperties>(inputStreamsNames.length);
			for (String streamName : inputStreamsNames) {			
				this.inputStreamsSchema.put(streamName, client.getStreamProperties(streamName)); 			
			}
		}			
		catch (Exception e) {				
			this.status.setStep(Step.ERROR);
			throw new Exception("Exception while retrieving streams list from server. \n(" + e.getMessage() + ")");				
		}					
	}
	
	protected void loadOutput(HashMap<String, InetSocketAddress> outputToMonitor) throws Exception {
		this.status.setStep(Step.LOADING);
		this.outputListeners = new StreamBaseListener[outputToMonitor.size()];	
		int i = 0;
		for (Entry<String, InetSocketAddress> e : outputToMonitor.entrySet()) {
			try {
				this.outputListeners[i] = new StreamBaseListener(this.sbAppURI, e.getKey(), e.getValue().getAddress().getHostAddress(), e.getValue().getPort());
				this.outputListeners[i].load();
				i++;
			}
			catch(Exception exc) {;
				this.status.setStep(Step.ERROR);
				throw new Exception("Could not load output listener for Monitor at " + 
						e.getValue().getAddress().getHostAddress() + ":" +
						e.getValue().getPort() +
						"\n(" + exc.getMessage() + ")");
			}
		}

		this.startAllEventListeners();
		
		this.status.setStep(Step.READY);
	}
	
	@Override
	public boolean load(HashMap<String, InetSocketAddress> outputToMonitor) throws Exception {
		// This interface instance has already been loaded
		if(this.status.getStep() == Step.READY) {
			return true;
		}
		else { // If it is not connected yet, try to connect
			if(this.status.getStep() != Step.CONNECTED) {
				this.connect();
			}
		}		
		
		if(this.status.getStep() == Step.CONNECTED) {			
			this.status.setStep(Step.LOADING);
			// Input
			loadInput();
			
			// Output
			loadOutput(outputToMonitor);
			
			return true;
		}
		else {
			return false;
		}
	}
	

	public String[] getInputStreamList() throws Exception {
		String ret[]=null;
		if(this.client != null) {
			ret = this.client.listEntities(EntityType.INPUT_STREAMS);
		}			
		
		return ret;
			
	}
	
	public String[] getOutputStreamList() throws Exception {
		String ret[]=null;
		if(this.status.getStep() == Step.CONNECTED && this.client != null) {
			ret = this.client.listEntities(EntityType.OUTPUT_STREAMS);
		}			
		
		return ret;
	}
	
	private void startAllEventListeners(){
		for (int i = 0; i < this.outputListeners.length; i++) {
			outputListeners[i].start();
		}
	}
	
	public void disconnect() {
		if (this.client != null) {
			try {
				this.client.close();
			} catch (StreamBaseException e) {
				System.err.println("Exception occurred while closing client: " + e.getMessage());
				e.printStackTrace();
			}
		}
		
		long totalReceivedEvents = 0;
		for (int i = 0; i < this.outputListeners.length; i++) {			
			outputListeners[i].close();		
			totalReceivedEvents += this.outputListeners[i].receivedEventCount;
		}				
	}		

	public void sendEvent(String e) throws Exception {
		if(this.status.getStep() == Step.READY || this.status.getStep() == Step.CONNECTED) {
			String ev[] = e.split(Globals.CSV_SEPARATOR);
			String destStream = ev[0].substring(5);			
			StreamProperties streamProp = this.inputStreamsSchema.get(destStream);
			if(streamProp != null) {			
				Tuple t = streamProp.getSchema().createTuple();
				
				// Set attributes' values of the event
				try {
					for (int i = 1; i < ev.length; i++) {
						t.setField(i-1, ev[i]);
					}
				}
				catch (TupleException e1) {
					throw new Exception("Could not send event. (" + e1.getMessage()+")");					
				}
								
				// sends event				
				client.enqueue(streamProp, t);
			}
			else {
				System.err.println("Unknown event type. It is not possible to send event.");
			}
		}
		else {
			System.err.println("Could not send event. The interface with StreamBase is not ready");
		}
	}

}
