package adapters;

import java.io.File;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.Socket;

import basic.Globals;

import com.streambase.sb.*;
import com.streambase.sb.client.*;

/**
 * Thread for listening incoming events from an output stream at StreamBase CEP Engine. 
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public class StreamBaseListener extends Thread{

	
	private StreamBaseClient client = null;
	private String streamName;		
	
	private String sinkAddress;
	private int sinkPort;
	private Socket sinkSocket;
	ObjectOutputStream sinkOutput;
	private Schema schema;
		
	public long receivedEventCount = 0; 
	private boolean keepListening = true;


	public StreamBaseListener(String serverURL, String streamName, String sinkAddress, int sinkPort) throws StreamBaseException {				
		this.streamName = streamName;
		this.sinkAddress = sinkAddress;
		this.sinkPort = sinkPort;
		/* Following StreamBase recommendations, one client is created for each listener:
		 * 	"If multiple threads subscribe to streams, 
		 * 	it is recommended that they use separate instances of this class."
		 */		
		this.client = new StreamBaseClient(serverURL);
		
		String logPath = Globals.APP_PATH + "log";
		File f = new File(logPath);
		if(!f.exists())
			f.mkdirs();
	}

	
	/**
	 * 
	 * Subscribes to an output stream on StreamBase
	 * 
	 * @throws StreamBaseException
	 */
	public void load() throws StreamBaseException {		
		// Subscribe to input stream
		client.subscribe(this.streamName);
	}

	public void run() {
		// Receive incoming events	
		DequeueResult dr = null;
		try {
			//"dequeue" method blocks until either data has arrived for a single stream,
			//or the client connection has been broken (returning null)
			while ((dr = client.dequeue()) != null && keepListening) {					
				for (Tuple tuple : dr) {
					this.processIncomingEvent(tuple);
				}

			}
		}
		 catch (StreamBaseException e) {
			System.err.println("Error while listening to events. ("+ e.getMessage() + ")");			
		}
		 catch(Exception e) {			 
			 System.err.println("Unrecoverable exception: dequeuing thread is now exiting.");
			 disconnect();
		 }
	}
	
	/**
	 * Process events from the Output streams:
	 * Translates the event from the CEP engine native format to the framework's CSV format and forwards to Sink
	 * 
	 * @param m		The incoming message
	 */
	public void processIncomingEvent(Tuple t) {				
		if(this.schema == null)
			this.schema = t.getSchema();
		
		String e = "type:" + streamName + Globals.CSV_SEPARATOR + t.toString();		

		// Forward event to Sink		
		forward(e);
	}
	
	private void forward(String e) {
		// If it is the first event, open connection to Sink		
		if(sinkSocket == null || sinkOutput == null || sinkSocket.isClosed()) {			
			try {
				connectToSink();
			} catch (IOException ioe1) {
				System.err.println("Could not connect to Sink (" + ioe1.getMessage() + "). Trying reconnection...");				
				forward(e);
			}		
		}
		
		try {			
			sinkOutput.writeObject(e);			
			this.receivedEventCount++;
			if(this.receivedEventCount%100000==0)
				sinkOutput.reset();
		} catch (IOException ioe) {
			System.err.println("Error while forwarding event to Sink. ("+ ioe.getMessage() + "). Trying reconnection...");			
			try {
				connectToSink();
				forward(e);
			} catch (IOException ioe1) {
				System.err.println("Sink seems to be offline.");
				
			}		
		}		
	}

	
	/**
	 * Stops listening for events
	 */
	public void close() {
		this.keepListening = false;
	}
	
	private void connectToSink() throws IOException {
		disconnectFromSink();
		this.sinkSocket = new Socket(sinkAddress, sinkPort);
		this.sinkSocket.setTcpNoDelay(true);
		this.sinkSocket.setSoLinger(true, 5);
		this.sinkOutput = new ObjectOutputStream(sinkSocket.getOutputStream());	
	}
	
	private void disconnectFromSink() {
		if(this.sinkOutput != null) {
			try {
				sinkOutput.close();
			} catch (IOException e) {
				System.err.println("Error while closing connection with Sink (" + e.getMessage() + ")");
			}
			sinkOutput = null;
		}
		
		if(this.sinkSocket != null) {
			try {
				sinkSocket.close();
			} catch (IOException e) {
				System.err.println("Error while closing connection with Sink (" + e.getMessage() + ")");
			}
			sinkSocket = null;
		}
	}
	
	
	/**
	 * Unsubscribes from output stream
	 * Closes the communication with the Sink
	 * 
	 * @return
	 */
	private void disconnect() {
		if (this.client != null) {
			try {
				this.client.close();
			} catch (StreamBaseException e) {
				System.err.println("StreamBaseException while closing client: " + e.getMessage());				
			}
		}
		
		disconnectFromSink();
				
	}
	

}
