package communication;

import java.io.File;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;

import data.CSVWriter;

import basic.Event;
import basic.Globals;

/**
 * Class used to encapsulate client-side communication between event-sending applications (Drivers/Monitors) 
 * and event-receiving applications (Adapters and Validators)
 * 
 * 
 * @author mnunes
 *
 */
public class ClientSocketInterface {
	private Socket s;
	private ObjectOutputStream output;	
	private int  eventSentCount=0;
	private CSVWriter logger;
	private boolean mustLog;
	
	/**
	 * 
	 * @param destinationAddress		The address where the receiving application is running
	 * @param destinationPort			The port to which the receiving application is listening
	 * @param log						Indicates if sent events must be logged to disk
	 * @param clientID					The ID of the sending application (used for logging purposes)
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	public ClientSocketInterface(InetAddress destinationAddress, int destinationPort,  boolean log, String clientID) throws UnknownHostException, IOException {
		this(destinationAddress.getHostAddress(), destinationPort, log, clientID );
	}
	

	/**
	 * 
	 * @param destinationAddress		The address where the receiving application is running
	 * @param destinationPort			The port to which the receiving application is listening
	 * @param log						Indicates if sent events must be logged to disk
	 * @param clientID					The ID of the sending application (used for logging purposes)
	 * @throws UnknownHostException
	 * @throws IOException
	 */
	public ClientSocketInterface(String destinationAddress, int destinationPort, boolean log, String clientID) throws UnknownHostException, IOException {
		// Socket communication
		s = new Socket(destinationAddress, destinationPort);	
		s.setTcpNoDelay(true);
		s.setSoLinger(true, 5);
		output = new ObjectOutputStream(s.getOutputStream());
		this.mustLog = log;
		
		if(log) {
			// Logging
			String logPath = Globals.APP_PATH + "log";
			File f = new File(logPath);
			if(!f.exists())
				f.mkdirs();
			logger = new CSVWriter(logPath + File.separator + clientID + ".log");
			
		}
	}
	
	/**
	 * Sends an event to another application (CEP adapter or Validator). It is a blocking operation
	 * 
	 * @param e			The event to be sent
	 * @throws IOException 
	 * 
	 */
	public synchronized void sendEvent(Event e) throws IOException {				
		long timestamp = System.currentTimeMillis();
		output.writeObject(e);			
		this.eventSentCount++;
		if(this.eventSentCount%100000==0)
			output.reset();
		if(mustLog)
			logger.writeRecord(e,timestamp);
	}
		
	
	/**
	 * Sends an event in String CSV format to another application (CEP adapter or Validator). 
	 * It is a blocking operation
	 *  
	 * @param event			The event to be sent
	 * @throws IOException 
	 * 
	 */	
	public synchronized void sendStringEvent(String event) throws IOException {		
		long timestamp = System.currentTimeMillis();
		output.writeObject(event);			
		this.eventSentCount++;
		if(this.eventSentCount%100000==0)
			output.reset();
		if(mustLog)
			logger.writeRecord(event, timestamp);
	}
	
	/**
	 * Sends an event in String CSV format to another application (CEP adapter or Validator), along with its timestamp. 
	 * Typically used to send events to Validator.
	 * It is a blocking operation.
	 * 
	 * @param event			The event to be sent
	 * @throws IOException 
	 * 
	 */	
	public synchronized void sendStringEventWithTimestamp(String event) throws IOException {		
		long timestamp = System.currentTimeMillis();
		output.writeObject(timestamp+Globals.CSV_SEPARATOR+event);			
		this.eventSentCount++;
		if(this.eventSentCount%100000==0)
			output.reset();
		if(mustLog)
			logger.writeRecord(event, timestamp);
	}
	
	
	/**
	 * Sends an event in String CSV format to another application (CEP adapter or Validator), along with its timestamp. 
	 * Typically used to send events to Validator.
	 * It is a blocking operation.
	 * 
	 * @param event			The event to be sent
	 * @param timestamp		The event's timestamp
	 * @throws IOException
	 */
	public synchronized void sendStringEventWithTimestamp(String event, long timestamp) throws IOException {				
		output.writeObject(timestamp+Globals.CSV_SEPARATOR+event);			
		this.eventSentCount++;
		if(this.eventSentCount%100000==0)
			output.reset();
		if(mustLog)
			logger.writeRecord(event, timestamp);
	}
	
	/**
	 * Closes the connection with the destination and the log file
	 * 
	 * @throws IOException
	 */
	public synchronized void disconnect() throws IOException  {
		if(logger != null) {
			this.logger.closeFile();	
		}		
		
		if(output != null) {
			output.close();
			output = null;
		}
		
		if(s != null) {
			s.close();
			s = null;
		}
	}
	
	public synchronized boolean isConnected () {
		return this.s.isConnected();
	}
}