package controller;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.io.StringWriter;
import java.net.InetAddress;
import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Map.Entry;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.DOMException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

import random.ConstantVariate;
import random.RandomExponentialVariate;
import random.RandomNormalVariate;
import random.RandomUniformVariate;
import random.Variate;
import sink.SinkConfig;
import basic.Attribute;
import basic.Datatype;
import basic.Domain;
import basic.EventType;
import basic.PredefinedListDomain;
import basic.RandomDomain;
import basic.SequentialDomain;
import driver.DriverConfig;
import driver.ExternalFileWorkloadPhase;
import driver.SyntheticWorkloadPhase;
import driver.WorkloadPhase;

/**
 * Class used to load and save configuration files containing tests setup
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public class ConfigurationParser {
	private Element xmlFileRoot;
	private File configFile;

	/**
	 * Opens a XML file containing test setup
	 * 
	 * @param path			Path to configuration file
	 * @throws ParserConfigurationException
	 * @throws SAXException
	 * @throws IOException
	 */
	public void open(String path) throws ParserConfigurationException, SAXException, IOException {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder;
		builder = factory.newDocumentBuilder();
		configFile = new File(path);
		Document doc = builder.parse(configFile);		
		this.xmlFileRoot = doc.getDocumentElement();
	}
	
	/**
	 * Saves a test setup into a new configuration file 
	 * 
	 * @param drivers			List of configured Drivers
	 * @param sinks				List of configured Sinks
	 * @param path				Path of the new configuration file to be saved
	 * @throws ParserConfigurationException
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws TransformerException
	 */
	public void saveAs(DriverConfig[] drivers, SinkConfig[] sinks, String path) throws ParserConfigurationException, FileNotFoundException, IOException, TransformerException {
		this.configFile = new File(path);
		this.save(drivers, sinks);
	}
	
	/**
	 * Saves a test setup into the current configuration file 
	 * 
	 * @param drivers 			List of configured Drivers
	 * @param sinks				List of configured Sinks
	 * @throws ParserConfigurationException
	 * @throws FileNotFoundException
	 * @throws IOException
	 * @throws TransformerException
	 */
	public void save(DriverConfig[] drivers, SinkConfig[] sinks) throws ParserConfigurationException, FileNotFoundException, IOException, TransformerException {
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder;
		builder = factory.newDocumentBuilder();
		Document doc = builder.newDocument();
		
		Element root = doc.createElement("FINCoS");
		doc.appendChild(root);
		
		Element driverList = doc.createElement("Drivers");
		Element sinkList = doc.createElement("Sinks");

		root.appendChild(driverList);
		root.appendChild(sinkList);		
		
		Element driver;
		for (DriverConfig dr : drivers) {
			driver = doc.createElement("Driver");			
			driver.setAttribute("name", dr.getAlias());
			driver.setAttribute("address", dr.getAddress().getHostAddress());
			driver.setAttribute("sendOutOfOrder", ""+dr.isSendOutOfOrder());
			
			Element server = doc.createElement("Server");
			server.setAttribute("address", dr.getServerAddress().getHostAddress());
			server.setAttribute("port", ""+dr.getServerPort());
			driver.appendChild(server);
			
			Element workload, phase, schema;
			workload = doc.createElement("Workload");
			SyntheticWorkloadPhase syntheticPhase;
			ExternalFileWorkloadPhase externalFilePhase;
			for (WorkloadPhase w : dr.getWorkload()) {				
				if(w instanceof SyntheticWorkloadPhase) {
					syntheticPhase = (SyntheticWorkloadPhase) w;
					phase = doc.createElement("Phase");
					phase.setAttribute("type", "Synthetic");
					Element duration = doc.createElement("duration");
					duration.appendChild(doc.createTextNode(""+syntheticPhase.getDuration()));
					Element initialRate = doc.createElement("initialRate");
					initialRate.appendChild(doc.createTextNode(""+syntheticPhase.getInitialRate()));
					Element finalRate = doc.createElement("finalRate");
					finalRate.appendChild(doc.createTextNode(""+syntheticPhase.getFinalRate()));
					phase.appendChild(duration);
					phase.appendChild(initialRate);
					phase.appendChild(finalRate);
					schema = this.saveSchema(doc, syntheticPhase.getSchema());
					phase.appendChild(schema);
					workload.appendChild(phase);	
				}
				else if(w instanceof ExternalFileWorkloadPhase) {
					externalFilePhase = (ExternalFileWorkloadPhase) w;	
					phase = doc.createElement("Phase");
					phase.setAttribute("type", "External File");
					Element path = doc.createElement("path");
					path.appendChild(doc.createTextNode(externalFilePhase.getFilePath()));
					Element timestamps = doc.createElement("timestamps");
					boolean containsTS = externalFilePhase.containsTimestamps();
					boolean useTS = externalFilePhase.isUsingTimestamps();
					int timeUnit = externalFilePhase.getTimestampUnit();
					double eventRate = externalFilePhase.getEventSubmissionRate();
					timestamps.setAttribute("contains", ""+containsTS);
					if(containsTS) {
						timestamps.setAttribute("use", ""+useTS);
						if(useTS) {
							timestamps.setAttribute("timeUnit", ""+timeUnit);
						}
						else {
							timestamps.setAttribute("eventRate", ""+eventRate);
						}
					}
					else {
						timestamps.setAttribute("eventRate", ""+eventRate);
					}					
					Element eventTypes = doc.createElement("eventTypes");
					boolean containsEventTypes = externalFilePhase.containsEventTypes();
					String singleTypeName = externalFilePhase.getSingleEventTypeName();
					eventTypes.setAttribute("contains", ""+containsEventTypes);
					if(!containsEventTypes)						
						eventTypes.setAttribute("singleTypeName", singleTypeName);
					
					phase.appendChild(path);
					phase.appendChild(timestamps);
					phase.appendChild(eventTypes);
					workload.appendChild(phase);
				}
				
			}
			driver.appendChild(workload);
			
			Element validation = doc.createElement("Validation");
			validation.setAttribute("validate", ""+dr.isValidationEnabled());
			if(dr.isValidationEnabled()) {
				validation.setAttribute("validatorAddress", dr.getValidatorAddress().getHostAddress());
				validation.setAttribute("validatorPort", ""+dr.getValidatorPort());
				validation.setAttribute("samplingRate", ""+dr.getValidationSamplingRate());	
			}
			
			driver.appendChild(validation);
			
			driverList.appendChild(driver);
		}
		
		Element sink;
		
		for (SinkConfig sinkCfg : sinks) {
			sink = doc.createElement("Sink");
			sink.setAttribute("name", sinkCfg.getAlias());
			sink.setAttribute("address", sinkCfg.getAddress().getHostAddress());
			sink.setAttribute("port", sinkCfg.getPort()+"");			
			Element server = doc.createElement("Server");
			server.setAttribute("address", sinkCfg.getServerAddress().getHostAddress());
			sink.appendChild(server);
			
			Element stream;
			for(String streamName : sinkCfg.getOutputStreamList()) {
				stream = doc.createElement("Stream");				
				stream.appendChild(doc.createTextNode(streamName));	
				sink.appendChild(stream);
			}
			sinkList.appendChild(sink);
			
			Element validation = doc.createElement("Validation");
			validation.setAttribute("validate", ""+sinkCfg.isValidationEnabled());
			if(sinkCfg.isValidationEnabled()) {
				validation.setAttribute("validatorAddress", sinkCfg.getValidatorAddress().getHostAddress());
				validation.setAttribute("validatorPort", ""+sinkCfg.getValidatorPort());
				validation.setAttribute("samplingRate", ""+sinkCfg.getValidationSamplingRate());	
			}
			
			sink.appendChild(validation);
		}
		
		BufferedWriter bw = new BufferedWriter(new FileWriter(this.configFile, false));
		bw.write(fromXMLDocToString(doc));
		bw.flush();
		bw.close();
		this.xmlFileRoot = doc.getDocumentElement();		
	}	
	
	/**
	 * Indicates if the parser already has a configuration file open
 
	 * @return
	 */
	public boolean isFileOpen() {
		return (this.xmlFileRoot != null);
	}
	
	/**
	 * Unloads parser's configuration file
	 * 
	 */
	public void closeFile() {
		this.xmlFileRoot = null;
	}
	
	/**
	 * Retrieves the list of Drivers in this parser's configuration file
	 * 
	 * @return			A list of configurations of Drivers
	 * @throws Exception 
	 * @throws DOMException 
	 * @throws NumberFormatException 
	 */
	public DriverConfig[] retrieveDriverList() throws NumberFormatException, DOMException, Exception  {
		DriverConfig ret[] = null;
		
		if(isFileOpen()) {
			Element driversList =(Element)this.xmlFileRoot.getElementsByTagName("Drivers").item(0);

			NodeList drivers = driversList.getElementsByTagName("Driver");
			ret = new DriverConfig[drivers.getLength()];

			Element driver;
			String driverName;
			InetAddress driverAddress, serverAddress, validatorAddress=null;
			int serverPort, validatorPort=0;
			boolean sendOutOfOrder, validate;
			double validSamplingRate=0;

			// Iterates over driver list
			for (int i = 0; i < drivers.getLength(); i++) {
				driver = (Element)drivers.item(i);
				driverName = driver.getAttribute("name");
				driverAddress = InetAddress.getByName(driver.getAttribute("address"));
				sendOutOfOrder = Boolean.parseBoolean(driver.getAttribute("sendOutOfOrder"));
				Element workload = (Element) driver.getElementsByTagName("Workload").item(0);
				NodeList phases = workload.getElementsByTagName("Phase");
				WorkloadPhase wps []= new WorkloadPhase[phases.getLength()];
				Element phase;				
				// Iterates over phases of a driver
				for (int j = 0; j < phases.getLength(); j++) {
					phase = (Element)phases.item(j);
					wps[j] = processPhase(phase);
				}
				Element server = (Element) driver.getElementsByTagName("Server").item(0);
				serverAddress = InetAddress.getByName(server.getAttribute("address"));
				serverPort = Integer.parseInt(server.getAttribute("port"));
				Element validation = (Element) driver.getElementsByTagName("Validation").item(0);
				validate = Boolean.parseBoolean(validation.getAttribute("validate"));			
				if(validate) {					
					validatorAddress = InetAddress.getByName(validation.getAttribute("validatorAddress"));
					validatorPort = Integer.parseInt(validation.getAttribute("validatorPort"));
					validSamplingRate = Double.parseDouble(validation.getAttribute("samplingRate"));
				}
				
				ret[i] = new DriverConfig(driverName, driverAddress, wps, sendOutOfOrder, serverAddress, serverPort, validate, validatorAddress, validatorPort, validSamplingRate);
			}	
		}

		return ret;
	}
	
	/**
	 * Parses the configuration of a Driver's phase
	 * 
	 * @param phase			A XML element containing the configuration of a phase
	 * @return				Either an instance of <tt>SyntheticWorkloadPhase</tt> or of <tt>ExternalFileWorkloadPhase</tt>
	 * @throws NumberFormatException
	 * @throws DOMException
	 * @throws Exception
	 */
	private WorkloadPhase processPhase(Element phase) throws NumberFormatException, DOMException, Exception {
		String type = phase.getAttribute("type");
		WorkloadPhase ret = null;
		
		if(type.equalsIgnoreCase("Synthetic")) {
			int phaseDuration;
			double phaseInitialRate, phaseFinalRate;
			
			phaseDuration = Integer.parseInt(
					phase.getElementsByTagName("duration").item(0).getFirstChild().getNodeValue());
			phaseInitialRate = Double.parseDouble(
					phase.getElementsByTagName("initialRate").item(0).getFirstChild().getNodeValue());
			phaseFinalRate = Double.parseDouble(
					phase.getElementsByTagName("finalRate").item(0).getFirstChild().getNodeValue());
			Element schema = (Element)phase.getElementsByTagName("Schema").item(0);
			ret = new SyntheticWorkloadPhase(phaseDuration, phaseInitialRate, phaseFinalRate, this.getSchema(schema));
		}
		else if(type.equalsIgnoreCase("External File")) {
			String path = phase.getElementsByTagName("path").item(0).getFirstChild().getNodeValue();
			Element timestamps = (Element) phase.getElementsByTagName("timestamps").item(0);			
			boolean containsTS = Boolean.parseBoolean(timestamps.getAttribute("contains"));
			boolean useTS = Boolean.parseBoolean(timestamps.getAttribute("use"));
			int timeUnit = 0;
			double eventRate = 1;
			if(containsTS && useTS)
				timeUnit = Integer.parseInt(timestamps.getAttribute("timeUnit"));
			else {
				eventRate = Double.parseDouble(timestamps.getAttribute("eventRate"));
			}
			Element eventTypes = (Element) phase.getElementsByTagName("eventTypes").item(0);
			boolean containsTypes = Boolean.parseBoolean(eventTypes.getAttribute("contains"));
			String singleTypeName = eventTypes.getAttribute("singleTypeName");			
			ret = new ExternalFileWorkloadPhase(path, containsTS, useTS, timeUnit, eventRate, containsTypes, singleTypeName);
		}
		else throw new Exception();
		
		return ret;
	}

	/**
	 * Retrieves the list of Sinks in this parser's configuration file
	 * 
	 * @return			A list of configurations of Sinks
	 * @throws Exception
	 */
	public SinkConfig[] retrieveSinkList() throws Exception {
		SinkConfig ret[] = null;
		
		if(isFileOpen()) {
			Element sinksList =(Element)this.xmlFileRoot.getElementsByTagName("Sinks").item(0);

			NodeList sinks = sinksList.getElementsByTagName("Sink");
			ret = new SinkConfig[sinks.getLength()];

			Element sink;
			String sinkName;
			InetAddress sinkAddress, serverAddress, validatorAddress=null;
			int sinkPort, validatorPort=0;
			boolean validate;
			double validSamplingRate=0;

			// Iterates over driver list
			for (int i = 0; i < sinks.getLength(); i++) {
				sink = (Element)sinks.item(i);
				sinkName = sink.getAttribute("name");
				sinkPort = Integer.parseInt(sink.getAttribute("port"));
				sinkAddress = InetAddress.getByName(sink.getAttribute("address"));			
				NodeList streams = sink.getElementsByTagName("Stream");			
				String streamNames []= new String[streams.getLength()];
				
				// Iterates over streams of a Sink
				for (int j = 0; j < streams.getLength(); j++) {
					streamNames[j] = streams.item(j).getFirstChild().getNodeValue();
				}
				
				Element server = (Element) sink.getElementsByTagName("Server").item(0);
				serverAddress = InetAddress.getByName(server.getAttribute("address"));				
				
				Element validation = (Element) sink.getElementsByTagName("Validation").item(0);
				validate = Boolean.parseBoolean(validation.getAttribute("validate"));			
				if(validate) {					
					validatorAddress = InetAddress.getByName(validation.getAttribute("validatorAddress"));
					validatorPort = Integer.parseInt(validation.getAttribute("validatorPort"));
					validSamplingRate = Double.parseDouble(validation.getAttribute("samplingRate"));
				}
				
				ret[i] = new SinkConfig(sinkName, sinkAddress, sinkPort, streamNames,
						serverAddress, validate, validatorAddress, validatorPort, validSamplingRate);
			}
			
			
		}
		return ret;
	}
	
	
	/**
	 * 
	 * Used to retrieve event types and their mixes of an execution phase of a given Driver
	 * 
	 * @param phaseSchema		An XML element containing the schema of a phase
	 * @return					A map <<tt>EventType</tt>, Double>, containing event mix
	 * @throws Exception 
	 * @throws DOMException 
	 * @throws NumberFormatException 
	 */
	private LinkedHashMap<EventType, Double> getSchema(Element phaseSchema) throws NumberFormatException, DOMException, Exception{
		LinkedHashMap<EventType, Double> ret=null;
		
		String typeName;
		Double typeMix;	
		
		NodeList typeList = phaseSchema.getElementsByTagName("EventType");
		ret = new LinkedHashMap<EventType, Double>(typeList.getLength());
		
		Element type;
		for (int i = 0; i < typeList.getLength(); i++) {
			type = (Element)typeList.item(i);
			typeName = type.getAttribute("name");
			typeMix = Double.parseDouble(type.getAttribute("mix"));
			NodeList atts = type.getElementsByTagName("Attribute");
			Attribute attributes[] = new Attribute[atts.getLength()];
			
			Element att;
			for (int j = 0; j < atts.getLength(); j++) {
				att = (Element)atts.item(j);
				String attName = att.getAttribute("name");
				Datatype attDataType = this.parseDataType(att.getAttribute("type"));
				Domain attDomain = this.parseDomain((Element)att.getElementsByTagName("Domain").item(0));
				attributes[j] = new Attribute(attDataType, attName, attDomain);
			}
			ret.put(new EventType(typeName, attributes), typeMix) ;
			
		}
		return ret;
	}
	
	private Datatype parseDataType(String datatype) {
		if (datatype.equalsIgnoreCase("INTEGER"))
			return Datatype.INTEGER;
		else {
			if (datatype.equalsIgnoreCase("LONG"))
				return Datatype.LONG;
			else {
				if (datatype.equalsIgnoreCase("FLOAT"))
					return Datatype.FLOAT;
				else {
					if (datatype.equalsIgnoreCase("DOUBLE"))
						return Datatype.DOUBLE;
						else {
							if (datatype.equalsIgnoreCase("TEXT"))
								return Datatype.TEXT;
							else {
								if (datatype.equalsIgnoreCase("BOOLEAN"))
									return Datatype.BOOLEAN;
								else
									return null;
								}
						}
					
					} 
				}
			}
	}
	
	/**
	 * Parses a Domain. Calls specific parsing methods for each kind of domain.
	 * 
	 * @param domain		A XML element containing domain's parameters
	 * @return				An instance of <tt>Domain</tt>
	 * @throws Exception 
	 * @throws DOMException 
	 * @throws NumberFormatException 
	 * @see 				<tt>Domain</tt> class
	 */
	private Domain parseDomain(Element domain) throws NumberFormatException, DOMException, Exception {
		Domain ret=null;
		
		String type = domain.getAttribute("type");
		
		if(type.equalsIgnoreCase("RANDOM"))
			ret = new RandomDomain(this.parseRandomVariate((Element)domain.getElementsByTagName("RandomVariable").item(0)));
		else {
			if(type.equalsIgnoreCase("SEQUENTIAL")) {
				ret = this.parseSequentialDomain(domain);
			}
			else {
				if (type.equalsIgnoreCase("PREDEFINED_LIST"))
					ret = this.parsePredefinedList(domain);
			}
		}
		
		return ret;
	}

	
	/**
	 * Parses a sequential Domain
	 * 
	 * @param domain		A XML element containing domain's parameters
	 * @return				An instance of <tt>SequentialDomain</tt>
	 * @throws Exception 
	 * @throws DOMException 
	 * @throws NumberFormatException 
	 * @see 				<tt>SequentialDomain</tt> class
	 */
	private Domain parseSequentialDomain(Element domain) throws NumberFormatException, DOMException, Exception{
		Domain ret;
		
		Variate initialValueVariate = null;
		Variate incrementVariate = null;
		
		// Parses Initial Value
		Element initialValue = (Element)domain.getElementsByTagName("InitialValue").item(0);
		if (initialValue.getAttribute("type").equalsIgnoreCase("Constant")) {
			initialValueVariate = new ConstantVariate(Double.parseDouble(initialValue.getFirstChild().getNodeValue()));
		}
		else {
			if (initialValue.getAttribute("type").equalsIgnoreCase("Random"))
				initialValueVariate = this.parseRandomVariate((Element)initialValue.getElementsByTagName("RandomVariable").item(0));
		}	
		
		// Parses Increment
		Element increment = (Element)domain.getElementsByTagName("Increment").item(0);
		if (increment.getAttribute("type").equalsIgnoreCase("Constant")) {
			incrementVariate = new ConstantVariate(Double.parseDouble(increment.getFirstChild().getNodeValue()));
		}
		else {
			if (increment.getAttribute("type").equalsIgnoreCase("Random"))
				incrementVariate = this.parseRandomVariate((Element)increment.getElementsByTagName("RandomVariable").item(0));
		}	
		
		ret = new SequentialDomain(initialValueVariate, incrementVariate);
		
		return ret;
	}

	
	/**
	 * Parses a domain of the type predefined list 
	 * 
	 * @param domain		A XML element containing domain's parameters
	 * @return				An instance of <tt>PredefinedListDomain</tt>
	 * @see 				<tt>PredefinedListDomain</tt> class
	 */
	private Domain parsePredefinedList(Element domain) {
		NodeList list = domain.getChildNodes();
		Element item;
		
		String mode = domain.getAttribute("behavior");
		
		if (mode.equalsIgnoreCase("Deterministic")) {
			ArrayList<Object> itemList = new ArrayList<Object>();
			Object itemValue;
			for (int i = 0; i < list.getLength(); i++) {
				if (list.item(i).getNodeType() == Node.ELEMENT_NODE){
					item = (Element)list.item(i);
					itemValue = item.getAttribute("value");
					itemList.add(itemValue);
				}
			}
			String items[] = new String[itemList.size()];
			items = itemList.toArray(items);
			
			return new PredefinedListDomain(items);
		}
		else {
			if (mode.equalsIgnoreCase("Stochastic")) {
				LinkedHashMap<Object, Double> items = new LinkedHashMap<Object, Double>();
				String key;
				Double frequency;
				
				for (int i = 0; i < list.getLength(); i++) {
					if (list.item(i).getNodeType() == Node.ELEMENT_NODE){
						item = (Element)list.item(i);
						key = item.getAttribute("value");
						try {
							frequency = Double.parseDouble(item.getAttribute("frequency"));
							items.put(key, frequency);
						}
						catch (ClassCastException ce) {
							System.err.println("Invalid frequency for item.");
							frequency = null;
						}
					}
				}
				
				return new PredefinedListDomain(items);
			}
		}
		return null;
	}

	/**
	 * Parses a Variate from XML configuration file
	 * 
	 * @param randomVariate			a Variate object represented as XML element
	 * @return						an instance of <tt>Variate</tt> class
	 * @throws Exception 
	 * @throws DOMException 
	 * @throws NumberFormatException 
	 * @see							<tt>Variate</tt> class
	 */
	private Variate parseRandomVariate(Element randomVariate) throws NumberFormatException, DOMException, Exception {
		Variate rv = null;
		
		if(randomVariate.getAttribute("type").equalsIgnoreCase("UNIFORM")){
			Element lowerBound = (Element)randomVariate.getElementsByTagName("lower").item(0);
			Element upperBound = (Element)randomVariate.getElementsByTagName("upper").item(0);
			rv = new RandomUniformVariate(Double.parseDouble(lowerBound.getFirstChild().getNodeValue()), 
										  Double.parseDouble(upperBound.getFirstChild().getNodeValue()));
		}
		else {
			if(randomVariate.getAttribute("type").equalsIgnoreCase("NORMAL")){
				Element mean = (Element)randomVariate.getElementsByTagName("mean").item(0);
				Element stdev = (Element)randomVariate.getElementsByTagName("stdev").item(0);
				rv = new RandomNormalVariate(Double.parseDouble(mean.getFirstChild().getNodeValue()), 
						  Double.parseDouble(stdev.getFirstChild().getNodeValue()));
			}
			else {
				if(randomVariate.getAttribute("type").equalsIgnoreCase("EXPONENTIAL")){
					Element lambda = (Element)randomVariate.getElementsByTagName("lambda").item(0);
					rv = new RandomExponentialVariate(Double.parseDouble(lambda.getFirstChild().getNodeValue()));
				}
			}
		}

		return rv;
	}
	
	/**
	 * Converts the schema of a phase to XML
	 * 
	 * @param doc			used to create new elements
	 * @param eventMix		event types and frequencies on dataset
	 * @return				an XML Element representing the schema
	 */
	private Element saveSchema(Document doc, Map<EventType, Double> eventMix) {
		Element schema = doc.createElement("Schema");
		Element type;
		EventType t;
		for (Entry<EventType, Double> e : eventMix.entrySet()) {
			type = doc.createElement("EventType");
			t = e.getKey();
			type.setAttribute("name", t.getName());
			type.setAttribute("mix", e.getValue()+"");
			Element attribute, domain;
			for (Attribute att : t.getAttributes()) {
				attribute = doc.createElement("Attribute");
				attribute.setAttribute("name", att.getName());
				attribute.setAttribute("type", att.getType().toString());
				domain = this.saveDomain(doc, att.getDomain());
				attribute.appendChild(domain);
				type.appendChild(attribute);				
			}
			schema.appendChild(type);
		}
		
		return schema;
	}
	
	/**
	 * Converts an instance of the <tt>Domain</tt> class into XML textual representation
	 * and writes it into a XML document
	 * 
	 * @param doc		The XML document where the <tt>Domain</tt> instance must be written to
	 * @param d			The element where the <tt>Domain</tt> instance must be written to
	 * @return			The <tt>Domain</tt> instance
	 * @see				<tt>Domain</tt> class
	 */
	private Element saveDomain(Document doc, Domain d) {
		Element domain = doc.createElement("Domain");
		
		if (d instanceof SequentialDomain) {
			SequentialDomain seq = (SequentialDomain) d;
			domain.setAttribute("type", "SEQUENTIAL");
			Element initialValue = doc.createElement("InitialValue");
			this.saveVariate(doc, initialValue, seq.getInitialVariate());
			Element increment = doc.createElement("Increment");;
			this.saveVariate(doc, increment, seq.getIncrementVariate());
			domain.appendChild(initialValue);
			domain.appendChild(increment);
		}
		else if(d instanceof PredefinedListDomain){
			PredefinedListDomain  def = (PredefinedListDomain) d;
			domain.setAttribute("type", "PREDEFINED_LIST");
			if(def.isDeterministic()) {
				domain.setAttribute("behavior", "Deterministic");
				Element item;
				for (Object itemName : def.getItems()) {
					item = doc.createElement("Item");
					item.setAttribute("value", ""+itemName);
					item.setAttribute("frequency", "1");
					domain.appendChild(item);
				}
			}
			else {
				domain.setAttribute("behavior", "Stochastic");
				Element item;
				for (Entry<Object, Double> e : def.getItemMix().entrySet()) {
					item = doc.createElement("Item");
					item.setAttribute("value", ""+e.getKey());
					item.setAttribute("frequency", ""+(Double)e.getValue());
					domain.appendChild(item);
				}
			}
				
		} 
		else if (d instanceof RandomDomain) {
			RandomDomain rand = (RandomDomain)d;
			//domain.setAttribute("type", "RANDOM");			
			this.saveVariate(doc, domain, rand.getVariate());			
		}
		
		return domain;
	}
	
	/**
	 * Converts an instance of the <tt>Variate</tt> class into XML textual representation 
	 * and writes it into a XML document
	 * 
	 * @param doc		The XML document where the Variate instance must be written to
	 * @param e			The element where the Variate instance must be written to
	 * @param v			The Variate instance
	 * @see				<tt>Variate</tt> class
	 */
	private void saveVariate(Document doc, Element e, Variate v) {
		if(v instanceof ConstantVariate) {
			e.setAttribute("type", "Constant");
			String value = ""+((ConstantVariate)v).getValue();
			e.appendChild(doc.createTextNode(value));			
		}
		else 
		{
			Element randomVar = doc.createElement("RandomVariable");
			e.setAttribute("type", "Random");
			e.appendChild(randomVar);
			if(v instanceof RandomExponentialVariate){
				randomVar.setAttribute("type", "EXPONENTIAL");
				String lambdaValue = "" + ((RandomExponentialVariate)v).getLambda();
				Element lambda = doc.createElement("lambda");
				lambda.appendChild(doc.createTextNode(lambdaValue));
				randomVar.appendChild(lambda);
			}
			else if(v instanceof RandomNormalVariate){
				randomVar.setAttribute("type", "NORMAL");
				String meanValue = "" + ((RandomNormalVariate)v).getMean();
				Element mean = doc.createElement("mean");
				mean.appendChild(doc.createTextNode(meanValue));
				String stdevValue = "" + ((RandomNormalVariate)v).getStdev();
				Element stdev = doc.createElement("stdev");
				stdev.appendChild(doc.createTextNode(stdevValue));
				randomVar.appendChild(mean);
				randomVar.appendChild(stdev);
			}
			else if(v instanceof RandomUniformVariate){
				randomVar.setAttribute("type", "UNIFORM");
				String lwrValue = "" + ((RandomUniformVariate)v).getLower();
				Element lower = doc.createElement("lower");
				lower.appendChild(doc.createTextNode(lwrValue));
				String uprValue = "" + ((RandomUniformVariate)v).getUpper();
				Element upper = doc.createElement("upper");
				upper.appendChild(doc.createTextNode(uprValue));
				randomVar.appendChild(lower);
				randomVar.appendChild(upper);
			}
		}		
	}
	
	/**
	 * Reads a XML document and puts it into a String object
	 * 
	 * @param doc			The XML document
	 * @return				A textual representation of the XML document
	 * @throws TransformerException
	 */
	private String fromXMLDocToString(Document doc) throws TransformerException {
		TransformerFactory tFactory = TransformerFactory.newInstance();
		Transformer transformer = tFactory.newTransformer();

		DOMSource source = new DOMSource(doc);
		StringWriter sw=new StringWriter();
		StreamResult result = new StreamResult(sw);
		transformer.transform(source, result);
		return sw.toString();

	}
	
}
