package driver;

import java.io.IOException;
import java.util.LinkedHashMap;


import basic.Attribute;
import basic.Domain;
import basic.Event;
import basic.EventType;
import basic.PredefinedListDomain;
import data.CSVWriter;


public class DataGen {	

	private long generatedEvents=0;
	private long totalEventCount;
	private boolean keepGenerating = true;
	
	/**
	 * Generates events according to a synthetic workload specified in the configuration file
	 * 
	 * @param workload			The synthetic workload specification
	 * @param dataFilesDir		The directory where the data file(s) must be stored
	 * @param fileCount			The number of files into which data will be stored
	 * @throws IOException
	 */
	public void generateData(SyntheticWorkloadPhase workload, String dataFilesDir, int fileCount) throws IOException {						
		int testDuration = workload.getDuration();
		double initialRate = workload.getInitialRate();
		double finalRate = workload.getFinalRate();
		LinkedHashMap<EventType, Double> types = workload.getSchema();
		
		// Converts set of types to a format suitable for PredefinedListDomain (Object, Double)
		LinkedHashMap<Object, Double> objectTypes = new LinkedHashMap<Object, Double>(types.size());		
		for (EventType type : types.keySet()) {
			objectTypes.put(type, types.get(type));
		}
		
		// Creates a  Domain to choose types iteratively according to the mix specified in the configuration file 
		PredefinedListDomain typeChooser = new PredefinedListDomain(objectTypes);
		
		// Used to create data file(s)
		CSVWriter writers[];
				
		// Variables involved in data generation (the events, attributes, and their values) 
		EventType type;
		Event e;
		Attribute att;
		Domain d;		
		
		// Workload Parameters
		double avgRate = (initialRate+finalRate)/2;				
		totalEventCount = Math.round(testDuration * avgRate);			
				
		writers = new CSVWriter[fileCount];
		for (int i = 0; i < writers.length; i++) {
			writers[i] = new CSVWriter(dataFilesDir +"\\" + (i+1) + ".csv");
		}
		
		try {
			// Iterates over #events
			for (long j = 0; j < totalEventCount; j++) {
				if(keepGenerating) {
					// Choose a type according to the mix
					type = (EventType) typeChooser.generateValue();
					// Iterates over attributes
					e = new Event(type);
					for (int k = 0; k < type.getAttributes().length; k++) {
						att = type.getAttributes()[k];
						d = att.getDomain();
						if (d != null)
							e.setAttributeValue(att, d.generateValue());
					}
													
					writers[(int)(j%fileCount)].writeRecord(e);
					this.generatedEvents++;					
				}
				else
					break;
			}

		} catch (Exception e1) {			
			e1.printStackTrace();
		} finally {
			// close the files
			for (int i = 0; i < writers.length; i++) {
				writers[i].closeFile();
			}
		}
	}
	
	public void stopDataGeneration() {
		this.keepGenerating = false;

	}
	
	public long getGeneratedEventsCount() {
		return generatedEvents;
	}
	
	public double getProgress() {
		return 1.0*this.generatedEvents/this.totalEventCount;
	}
}
