package driver;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.text.DecimalFormat;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JProgressBar;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.Timer;
import javax.swing.border.EtchedBorder;

import basic.EventType;
import basic.Globals;
import basic.InvalidStateException;
import basic.Status;
import basic.Step;

import communication.ClientSocketInterface;

import data.DataFileReader;

@SuppressWarnings("serial")

/**
 *	Class responsible for load generation 
 */
public class Driver extends JFrame implements DriverRemoteFunctions{
	
	//Setup
	private static final String DEFAULT_DATA_FILES_DIR = Globals.APP_PATH + "data";;		
	private DataGen dg;
	private DriverConfig drConfig;
	private String alias;
	
	// Run
	private Sender senders[];
	private int threadCount; // Number of Sender threads (applies only for synthetic workloads)
	
	// Communication
	private ClientSocketInterface server;
	private ClientSocketInterface validator;
	
	// Status and Statistics
	private Status status;		
	private int totalEventCount = 0;
	private long totalEventsSent=0;
	private int phaseEventsSent=0;
	
		
//=========================== GUI ===================================
	JTextArea infoArea;
	JLabel statusLabel, progressLabel, eventCountLabel;	
	JPanel statusPanel;
	JProgressBar bar;
//===================================================================
	
	public Driver() {
		this("Driver");
	}
	
	public Driver(String alias) {
		super(alias);
		this.alias = alias;
		statusPanel =  new JPanel();
		statusPanel.setLayout(null);
        statusPanel.setPreferredSize(new Dimension(200,400));
        statusPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED), "Status"));
        statusLabel = new JLabel();        
        bar = new JProgressBar(10,100);
        bar.setStringPainted(true);
        progressLabel = new JLabel("Progress: ");
        eventCountLabel = new JLabel("Sent events: " + totalEventsSent);
        
		statusPanel.add(statusLabel);
		statusPanel.add(progressLabel);
		statusPanel.add(bar);
		statusPanel.add(eventCountLabel);
		
		statusLabel.setBounds(10, 10, 150, 50);
		progressLabel.setBounds(20, 60, 60, 15);
		bar.setBounds(80, 60, 100, 15);
		eventCountLabel.setBounds(20, 75, 150, 50);
        
		infoArea = new JTextArea(10, 30);   
        infoArea.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED), "Info"));
        JScrollPane infoScroll = new JScrollPane(infoArea);
        infoArea.setEditable(false);        
        
		this.getContentPane().setLayout(new BorderLayout());
		this.getContentPane().add(statusPanel, BorderLayout.LINE_START);
		this.getContentPane().add(infoScroll, BorderLayout.CENTER);
		
		this.setSize(700, 400);        
		this.setLocationRelativeTo(null); //screen center
		this.setResizable(false);
		this.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
		this.setVisible(true);
		
		this.status = new Status();	
        this.updateStatus(this.status.getStep(), this.status.getProgress());
        System.out.println("Driver application started. Initializing remote interface...");
		showInfo("Driver application started. Initializing remote interface...");
		
		try {			
			this.initializeRMI();
			showInfo("Done! Waiting for remote commands...");
		} catch (Exception e) {
			this.updateStatus(Step.ERROR, 0);			
			System.err.println("ERROR: Could not initialize remote interface: " + e.getMessage());
			showInfo("ERROR: Could not initialize remote interface: " + e.getMessage());			
			try {
				Thread.sleep(5000);
			} catch (InterruptedException e1) {
				e1.printStackTrace();
			}
			System.exit(0);
		} 
		
		int delay = 1000/Globals.DEFAULT_GUI_REFRESH_RATE;
		Timer guiRefresher = new Timer(delay, new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent e) {				
				// Refresh Status information				
				getStatus();
				//Refresh screen with update Status information 
				updateScreen();						
			}			
		});
		guiRefresher.start();
				
	}
	
	public void showInfo(String msg) {
		Date now = new Date();		
		infoArea.append(Globals.TIME_FORMAT.format(now) + " - " + msg + "\n" );
		infoArea.setCaretPosition(infoArea.getDocument().getLength());
	}
	
	private void initializeRMI() throws RemoteException {	
		DriverRemoteFunctions stub = (DriverRemoteFunctions) UnicastRemoteObject.exportObject(this, 0);
		Registry registry = LocateRegistry.getRegistry();
		registry.rebind(this.alias, stub);			
	}
	
	
	private void updateStatus(Step s, double progress){
		synchronized (this.status) {
			this.status.setStep(s);
			this.status.setProgress(progress);		
			updateScreen();	
		}
		
	}
	
	private void updateScreen() {
		switch (this.status.getStep()) {
		case DISCONNECTED:
			this.statusLabel.setIcon(Globals.YELLOW_SIGN);
			break;
		case CONNECTED:
		case READY:
		case PAUSED:
		case STOPPED:
		case FINISHED:
			this.statusLabel.setIcon(Globals.BLUE_SIGN);
			break;		
		case LOADING:
		case RUNNING: 
			this.statusLabel.setIcon(Globals.GREEN_SIGN);
			break;
		case ERROR:
			this.statusLabel.setIcon(Globals.RED_SIGN);
			break;
		}
		
		this.statusLabel.setText(this.status.getStep().toString());
		this.eventCountLabel.setText("Sent events: " + (totalEventsSent+phaseEventsSent));
		this.bar.setValue((int) (100*this.status.getProgress()));		
		
		statusPanel.revalidate();
	}
	
	public Status getStatus() {	
		synchronized (this.status) {
			switch (this.status.getStep()) {		
				case READY:			
				case LOADING:
					if(dg != null)
						this.status.setProgress(dg.getProgress());
					else
						this.status.setProgress(1.0);
					break;
				case RUNNING: 
				case PAUSED:
				case ERROR:
				{
					phaseEventsSent = 0;
					if(this.senders != null) {
						for (Sender sender: this.senders) {
							if(sender != null)
								phaseEventsSent+=sender.getSentEventCount();
						}
					}
						
					this.status.setProgress((1.0*(totalEventsSent+phaseEventsSent)/this.totalEventCount));				
					break;
				}
				case FINISHED: 
				case STOPPED:{
					phaseEventsSent = 0;
					this.status.setProgress((1.0*(totalEventsSent+phaseEventsSent)/this.totalEventCount));
				}
			}

			if(this.senders != null) {
				for (Sender sender: this.senders) {
					//If there is at least one sender thread with error:
					if(sender != null && sender.getStatus().getStep() == Step.ERROR) {
						if(this.status.getStep() != Step.ERROR) {
							this.status.setStep(Step.ERROR);
							showInfo("Error during load submission.");
						}								
						break;
					}
					//Else, if there is at least one sender thread running:
					else if(sender != null && sender.getStatus().getStep() == Step.RUNNING) {
						this.status.setStep(Step.RUNNING);
						break;
					}
				}
			}
		}

		
		return this.status;
	}
	
	@Override
	public boolean load(DriverConfig dr, String dataFilesDir) throws InvalidStateException {				
		if(this.status.getStep() == Step.DISCONNECTED ||
		   this.status.getStep() == Step.ERROR ||
		   this.status.getStep() == Step.FINISHED ||
		   this.status.getStep() == Step.STOPPED ||
		   this.status.getStep() == Step.CONNECTED) 
		{
			showInfo("Loading Driver...");			
			this.totalEventCount = 0;
			this.totalEventsSent = 0;
			
			// Tries to connect to adapter
			if(server != null) {
				try {
					server.disconnect();
					server = null;
				} catch (IOException e) {
					System.err.println("Could not disconnect from server. (" +e.getMessage() +")");					
				}
			}
								
			try {
				this.showInfo("Trying to establish connection to server at " + dr.getServerAddress().getHostAddress() + ":" + dr.getServerPort() + "...");
				this.server = new ClientSocketInterface(dr.getServerAddress(), dr.getServerPort(), true, dr.getAlias());			
				this.showInfo("Done!");
			} catch (UnknownHostException e) {
				this.showInfo("ERROR: Cannot connect to specified server: Unknown host.");
				this.updateStatus(Step.ERROR, this.status.getProgress());
				return false;
			} catch (IOException e) {
				this.showInfo("ERROR: Cannot connect to specified server. (" + e.getMessage() + ").");
				this.updateStatus(Step.ERROR, this.status.getProgress());			
				return false;
			}			
			
			this.drConfig = dr;
			
			if(dr.isValidationEnabled()) {
				// Tries to connect to validator
				if(validator != null) {
					try {
						validator.disconnect();
						validator = null;
					} catch (IOException e) {
						System.err.println("Could not disconnect from validator. (" +e.getMessage() +")");						
					}
				}
									
				try {
					this.showInfo("Trying to establish connection to validator at " + dr.getValidatorAddress().getHostAddress() + ":" + dr.getValidatorPort() + "...");
					this.validator = new ClientSocketInterface(dr.getValidatorAddress(), dr.getValidatorPort(), false, dr.getAlias());			
					this.showInfo("Done!");
				} catch (UnknownHostException e) {
					this.showInfo("ERROR: Cannot connect to specified validator: Unknown host.");
					this.updateStatus(Step.ERROR, this.status.getProgress());
					return false;
				} catch (IOException e) {
					this.showInfo("ERROR: Cannot connect to specified validator. (" + e.getMessage() + ").");
					this.updateStatus(Step.ERROR, this.status.getProgress());			
					return false;
				}
			}			
			this.updateStatus(Step.CONNECTED, 0);			
						
			int totalTestDuration = 0;
			int synthPhaseCount = 0;
			int extFilePhaseCount = 0;
			long phaseEventCount;
			long t0, t1;
			
			if(dr.isSendOutOfOrder())
				threadCount = 3;
			else
				threadCount = 1;
			 			
			showInfo("Initializing Workload ("+ dr.getWorkload().length + " phases)");
			t0 = System.currentTimeMillis();			
			for (int i = 0; i < dr.getWorkload().length; i++) {
				if(dr.getWorkload()[i] instanceof SyntheticWorkloadPhase) {
					synthPhaseCount++;
					SyntheticWorkloadPhase syntheticPhase = (SyntheticWorkloadPhase) dr.getWorkload()[i];
					showInfo(" Phase " + (i+1) + ": Synthetic workload"
							+ "\n\tApproximate phase duration: " + syntheticPhase.getDuration() + " seconds (" + new DecimalFormat("###.##").format(syntheticPhase.getDuration()/3600.00)+" hour(s))."
							+"  \n\tEvent submission rates:\n\t\tinitial: " + syntheticPhase.getInitialRate() + " events/second."+ 
																"\n\t\tfinal:" + syntheticPhase.getFinalRate() + " events/second."+ 
																"\n\t\taverage: " + ((syntheticPhase.getInitialRate()+syntheticPhase.getFinalRate())/2)+ " events/second."														
					);
								
					try {				
						showInfo("\tLoading data...");
						dg = new DataGen();
						this.updateStatus(Step.LOADING, 0);
						// Cleans data directory
						File phaseDataDir = new File(DEFAULT_DATA_FILES_DIR+File.separator+dr.getAlias()+File.separator+"phase_" + (i+1));
						if(phaseDataDir.exists()) {
							for (
									File  f : phaseDataDir.listFiles(new FileFilter(){
										public boolean accept(File pathname) {
											return pathname.getName().contains(".csv");
										}})
								) f.delete();						
						}	
						else {
							phaseDataDir.mkdirs();	
						}				
							
						// Generates events
						dg.generateData(syntheticPhase, phaseDataDir.getAbsolutePath(), threadCount);
						if(this.status.getStep() == Step.STOPPED) {
							return false;
						}
						
						totalTestDuration += syntheticPhase.getDuration();
						phaseEventCount = dg.getGeneratedEventsCount();
						totalEventCount +=  phaseEventCount;
						showInfo("\tDone! " + phaseEventCount + " events generated.");
					} catch (IOException ioe) {
						showInfo("ERROR: Could not load data (" + ioe.getMessage() + ")");
						this.updateStatus(Step.ERROR, this.status.getProgress());				
						return false;
					}
					
				}
				else if (dr.getWorkload()[i] instanceof ExternalFileWorkloadPhase){
					extFilePhaseCount++;
					ExternalFileWorkloadPhase filePhase = (ExternalFileWorkloadPhase)dr.getWorkload()[i];
					String detailInfo = " Phase " + (i+1) + ": External File workload"
										+ "\n\tFile path: " + filePhase.getFilePath();
					// timestamp info
					if(filePhase.containsTimestamps()) {
						detailInfo+= "\n\tContains timestamps: Yes, expressed in ";						
						switch (filePhase.getTimestampUnit()) {
						case ExternalFileWorkloadPhase.MILLISECONDS:
							detailInfo+= "milliseconds";
							break;
						case ExternalFileWorkloadPhase.SECONDS:
							detailInfo+= "seconds";
							break;
						case ExternalFileWorkloadPhase.DATE_TIME:
							detailInfo+= "date/time";				
						}						
						if(filePhase.isUsingTimestamps())
							detailInfo+= "\n\tUse timestamps for event submission: Yes";
						else {
							detailInfo+= "\n\tUse timestamps for event submission: No";
							detailInfo+= "\n\tEvents will be submitted at a rate of " + filePhase.getEventSubmissionRate() + " events/sec.";
						}							
					}
					else{
						detailInfo+= "\n\tContains timestamps: No";
						detailInfo+= "\n\tEvents will be submitted at a rate of " + filePhase.getEventSubmissionRate() + " events/sec.";
					}
					//Event types info
					if(filePhase.containsEventTypes()) {
						detailInfo+= "\n\tContains event types: Yes";
					}
					else {
						detailInfo+= "\n\tContains event types: No";
						detailInfo+= "All events in the file are of the same type: " + filePhase.getSingleEventTypeName();
					}
						
					showInfo(detailInfo);														
				}
			}
			t1 = System.currentTimeMillis();
			
			this.updateStatus(Step.READY, 100);
			showInfo(" Data loading finished (elapsed Time: " + (t1-t0)/1000 + " seconds)" +
					"\n\t # Synthetic phases: " + synthPhaseCount + " (Total of events generated: " + totalEventCount + ")" +
					"\n\t # External File phases: " + extFilePhaseCount + ".");				
			
			
			return true;	
		}
		else {
			showInfo("Cannot load driver. Driver has already been loaded.");
			throw new InvalidStateException("Cannot load driver. Driver has already been loaded.");				
		}
		
	}
	
	public void start() throws InvalidStateException {	
		if(this.status.getStep() == Step.READY) {			
			// Initializes a new thread to run load, so RMI thread will return immediately after call.
			new Thread () {
				
				@Override
				public void run() {
					try {
						ThreadGroup senderGroup;
						long testT0, phaseT0;
						Scheduler sch;
						DataFileReader reader;						
						testT0 = System.currentTimeMillis();
						
						for (int i = 0; i < drConfig.getWorkload().length; i++) {
							phaseT0 = System.currentTimeMillis();
							if(drConfig.getWorkload()[i] instanceof SyntheticWorkloadPhase) {
								double initialThreadRate, finalThreadRate;
								LinkedHashMap<EventType, Double> schema;
								Set<EventType> types=null;
								SyntheticWorkloadPhase syntheticPhase = (SyntheticWorkloadPhase)drConfig.getWorkload()[i];
								senderGroup = new ThreadGroup("Senders");
								showInfo("Phase " + (i+1) + " started. Initializing dispatcher threads (Thread Count: " + threadCount + ")");								
								schema = syntheticPhase.getSchema();
								if(schema != null)
									types = schema.keySet();
								initialThreadRate = syntheticPhase.getInitialRate()/threadCount;
								finalThreadRate  = syntheticPhase.getFinalRate()/threadCount;
								// Creates dispatcher threads		
								senders = new Sender[threadCount];

								for (int j = 0; j < threadCount; j++) {					
									sch = new Scheduler(initialThreadRate, finalThreadRate, syntheticPhase.getDuration());
									reader = new DataFileReader(DEFAULT_DATA_FILES_DIR+File.separator + drConfig.getAlias() + File.separator + "phase_" + (i+1)+File.separator +(j+1)+".csv", types);
									if(drConfig.isValidationEnabled()) {
										senders[j] = new Sender(server, validator, sch, reader, false, senderGroup, ""+(j+1), drConfig.getValidationSamplingRate());	
									} 
									else {
										senders[j] = new Sender(server, sch, reader, false, senderGroup, ""+(j+1));
									}

									senders[j].start();	
								}
								updateStatus(Step.RUNNING, 0);
								showInfo("Done. Sending events...");

								// Waits for completion of all sending threads	
								for (Sender sender : senders) {									
									sender.join();									
								}

								for (Sender sender: senders) {
									totalEventsSent+=sender.getSentEventCount();
									phaseEventsSent-=sender.getSentEventCount();
								}								

							}
							else if(drConfig.getWorkload()[i] instanceof ExternalFileWorkloadPhase) {
								showInfo("Phase " + (i+1) + " started. Initializing dispatcher thread...");
								ExternalFileWorkloadPhase filePhase = (ExternalFileWorkloadPhase) drConfig.getWorkload()[i];
								senders = new Sender[1];

								reader = new DataFileReader(filePhase.getFilePath(), filePhase.containsTimestamps(), filePhase.containsEventTypes(), filePhase.getSingleEventTypeName());
								// Event submission is Timestamp-based
								if(filePhase.containsTimestamps() && filePhase.isUsingTimestamps()) {
									if(drConfig.isValidationEnabled()) {
										senders[0] = new Sender(server, validator, reader, filePhase.getTimestampUnit(), drConfig.getValidationSamplingRate());	
									}
									else {
										senders[0] = new Sender(server, reader, filePhase.getTimestampUnit());
									}
									
								}
								else {// Event submission is scheduled based on a fixed rate									
									sch = new Scheduler(filePhase.getEventSubmissionRate(), filePhase.getEventSubmissionRate(), 1);									
									if(drConfig.isValidationEnabled()) {
										senders[0] = new Sender(server, validator, sch, reader, filePhase.containsTimestamps(), drConfig.getValidationSamplingRate());					
									}
									else {
										senders[0] = new Sender(server, sch, reader, filePhase.containsTimestamps());
									}
								}
								senders[0].start();	
								
								updateStatus(Step.RUNNING, 0);
								showInfo("Done! Sending events...");

								// Waits for completion of the sending thread																	
								senders[0].join();									
							}
							showInfo("Phase "  + (i+1) + " finished (elapsed time: " + ((System.currentTimeMillis()-phaseT0))/1000 +  " seconds).");
							// Checks if driver has been stopped
							synchronized (status) {
								if(status.getStep() == Step.STOPPED) {
									try {
										server.disconnect();	
									}
									catch (IOException ioe1){
										System.err.println("Could not disconnect from server. (" +ioe1.getMessage() +")");
										showInfo("Could not disconnect from server. (" +ioe1.getMessage() +")");
									}

									return;
								}	
							}
						}				

						showInfo("Test finished. Total test duration: " + ((System.currentTimeMillis()-testT0))/1000 +  " seconds).");						
						updateStatus(Step.FINISHED, status.getProgress());			
					}
					  catch (IOException ioe) {
						updateStatus(Step.ERROR, status.getProgress());
						showInfo("Exception: " + ioe.getMessage());
						ioe.printStackTrace();
					} catch (InterruptedException e) {
						updateStatus(Step.ERROR, status.getProgress());
						showInfo("Interrupted Exception");
						e.printStackTrace();
					} finally {
						try {
							server.disconnect();
						} catch (IOException e) {							
							showInfo("Could not disconnect from server. (" +e.getMessage() +")");
						}
					}
				}
			}.start();			
		}
		else {
			if(this.status.getStep() == Step.PAUSED)
				this.resume();
			else {
				showInfo("Cannot start driver. Driver has not been loaded.");
				throw new InvalidStateException("Driver has not been loaded.");	
			}					
		}			
	}
	
	private void resume() {
		for (Sender sender : this.senders) {
				sender.resumeLoad();
		}
		this.status.setStep(Step.RUNNING);
		showInfo("Driver has been resumed.");
	
	}
	
	public void pause() throws InvalidStateException {
		synchronized (this.status) {
			if(this.status.getStep() == Step.RUNNING) {
				for (Sender sender : this.senders) {
					sender.pauseLoad();
				}
				this.status.setStep(Step.PAUSED);
				showInfo("Driver is paused.");
			}
			else {
				showInfo("Cannot pause Driver. Driver must be running in order to be paused.");
				throw new InvalidStateException("Driver must be running in order to be paused.");
			}	
		}	
	}
	
	public void stop() throws InvalidStateException{
		synchronized (this.status) {
			if(this.status.getStep() == Step.RUNNING ||
			   this.status.getStep() == Step.PAUSED ||
			   this.status.getStep() == Step.ERROR) 
			{
				this.status.setStep(Step.STOPPED);
				if(this.senders != null) {
					for (Sender sender : this.senders) {
						sender.stopLoad();	
						sender = null;
					}			
					this.senders = null;	
				}
				
				showInfo("Driver has been stopped.");
			}
			else if(this.status.getStep() == Step.LOADING) {
				this.status.setStep(Step.STOPPED);
				this.dg.stopDataGeneration();			
				showInfo("Data generation was stopped.");
			}
			else if(this.status.getStep() == Step.READY || this.status.getStep() == Step.FINISHED) {
				this.status.setStep(Step.STOPPED);						
				showInfo("Driver has been stopped.");
			}
			else 
				throw new InvalidStateException("Driver must be loading, ready, running or paused in order to be stopped.");	
		}
				
	}

	@Override
	public void switchToNextPhase() throws RemoteException, InvalidStateException {
		synchronized (this.status) {
			if(this.status.getStep() == Step.RUNNING ||
			   this.status.getStep() == Step.PAUSED) 
			{
				showInfo("Switching to next phase");
				for (Sender sender : this.senders) {
					sender.stopLoad();
				}
				showInfo("Done!");
			}
			else 
				throw new InvalidStateException("Driver must be running or paused in order to switch of phase.");

		}
	}

	@Override
	public void alterRate(double factor) throws InvalidStateException, RemoteException {
		synchronized (this.status) {
			if(this.status.getStep() == Step.RUNNING ||
					this.status.getStep() == Step.PAUSED) 
			{
				showInfo("Changing event submission rate (factor: " + factor + "x)");
				for (Sender sender : this.senders) {
					sender.setRateFactor(factor);
				}
				showInfo("Done!");
			} 
			else
				throw new InvalidStateException("Driver must be running or paused in order to alter event submission rate.");			
		}		
	}	
	
}
