package driver;

import java.util.Random;


public class Scheduler {
	private double initialRate;	// events/nanosecond
	private double finalRate;	// events/nanosecond
	private long testDuration;     // nanoseconds
	private long currentTime = 0;  // nanoseconds
	private double rateFactor = 1.0; // used to control event submission speed
	
	private Random rnd;	
	
	 /**
	  * 
	  * @param initialEventRate	The initial rate of submission of events (events/sec)
	  * @param finalEventRate	The final rate of submission of events (events/sec)
	  * @param testDuration		Total test duration, in seconds
	  * @throws Exception
	  */
	 public Scheduler(double initialEventRate, double finalEventRate, long testDuration) {
		 this.setInitialRate(initialEventRate/1E9); 		// converts to events/nanoseconds
		 this.setFinalRate(finalEventRate/1E9);			// converts to events/nanoseconds
		 this.setTestDuration((long)(testDuration*1E9));	// converts to nanoseconds		 
		 this.rnd = new Random();			 
	}
	 
	 /**
	  * 
	  * @param eventRate		The rate of submission of events (events/sec)
	  * @param testDuration	Total test duration, in seconds
	  * @throws Exception	
	  */
	 public Scheduler(double eventRate, int testDuration) {
		 this(eventRate, eventRate, testDuration);			 
	 }

	private void setInitialRate(double rate) {
		if(rate > 0) 
			this.initialRate = rate;
		else 
			this.initialRate = 1E-10; // 0.1 events/second
	}
	
	private void setFinalRate(double finalRate)  {
		if(finalRate > 0) 
			this.finalRate = finalRate;
		else 
			this.finalRate = 1E-10; //0.1 events/second
	}
	

	private void setTestDuration(long testDuration) {
		if(testDuration > 0)	
			this.testDuration = testDuration;
		else
			this.testDuration = 1;
	}

	/**
	 * Gets interarrival times of events 
	 * Poisson Process, interarrival times exponentially distributed
	 * 
	 * @return	The time in nanoseconds
	 */
	public long getInterArrivalTime() {
	//	long interTime = Math.round(-1/this.getEventRate(this.currentTime)*Math.log(rnd.nextDouble()));
		long interTime = Math.round(1/this.getEventRate(this.currentTime));	
		this.currentTime = this.currentTime + interTime;
		return interTime;	
	}
	
	/**
	 * Gets relative timestamps for events 
	 * @return The timestamp in nanoseconds
	 */
	public long getNextTime() {
		this.getInterArrivalTime();
		return this.currentTime;
	}
	
	/**
	 * Computes current event rate (a linear growth is presupposed)
	 * 
	 * @param t		The current time, in nanoseconds
	 * @return		Current event rate, considering current time and the initial and final rates 
	 */
	private double getEventRate(long t) {
		return this.rateFactor*(this.initialRate + t*(this.finalRate-this.initialRate)/this.testDuration);
	}
	
	public long getInterArrivalTime2() {
		double p = 1/this.getEventRate(this.currentTime);
		long interTime = Math.round(-1/this.getEventRate(this.currentTime)*Math.log(rnd.nextDouble()));
		this.currentTime = this.currentTime + interTime;
		if (rnd.nextDouble() < p)
			return 1;
		else
			return 0;
	}
	
	/**
	 * Sets the factor by which the event rate is multiplied
	 * 
	 * @param factor
	 */
	public synchronized void setRateFactor(double factor) {
		this.rateFactor = factor;
	}

	public static void main(String[] args) {
		Scheduler sch = new Scheduler(100000, 100000);
		long interTime, cumulativeTime = 0;
		int sleepCount = 0;
		int totalSleepTime = 0;
		for (int i = 0; i < 100000; i++) {
			interTime = sch.getInterArrivalTime();
			cumulativeTime+=interTime;
			//System.out.println(interTime);
			if(cumulativeTime >= 1000000) {
				totalSleepTime+=cumulativeTime;
				cumulativeTime = 0;
				
				sleepCount++;
			}
			
				
			
		}
		
		System.out.println("Sleep time: " + totalSleepTime);
		System.out.println("Sleep count: " + sleepCount);
	}
}
