package driver;

import java.io.IOException;

import communication.ClientSocketInterface;

import data.DataFileReader;
import basic.Globals;
import basic.Status;
import basic.Step;

public class Sender extends Thread{	
	private Scheduler scheduler;
	private DataFileReader data;	
	private ClientSocketInterface server, validator; 
	private long sentEventCount = 0;
	private long timeInPause = 0;
	private Status status;
	private int samplMod ;	
	private boolean containsTimestamps;
	private int timestampUnit;
	
	/**
	 * Constructor for sending events for both server and validator
	 * Event submission is controlled by the Scheduler passed as argument
	 * 
	 * @param server					A socket interface to the server/adapter 
	 * @param validator					A socket interface to a validator
	 * @param scheduler					Schedules event submission according to a given rate
	 * @param dataReader				Reads data sets from disk
	 * @param containsTimestamps		Indicates if the events in the data file are associated to timestamps
	 * @param group						A thread group
	 * @param id						A thread ID
	 * @param validationSamplingRate	The percentage of events to forward to a validator
	 */
	public Sender(ClientSocketInterface server, ClientSocketInterface validator, Scheduler scheduler, DataFileReader dataReader, 
			boolean containsTimestamps, ThreadGroup group, String id, double validationSamplingRate) {
		super(group, id);
		this.server = server;
		this.validator = validator;
		this.scheduler = scheduler;
		this.data = dataReader;
		this.containsTimestamps = containsTimestamps;
		this.status = new Status(Step.STOPPED, 0);
		this.samplMod = (int) (1/validationSamplingRate);
	}
	
	/**
	 * Constructor for sending events only for server (no validation)
	 * Event submission is controlled by the Scheduler passed as argument
	 * 
	 * @param server				A socket interface to the server/adapter 
	 * @param scheduler				Schedules event submission according to a given rate
	 * @param dataReader			Reads data sets from disk
	 * @param containsTimestamps	Indicates if the events in the data file are associated to timestamps
	 * @param group					A thread group
	 * @param id					A thread ID
	 */
	public Sender(ClientSocketInterface server, Scheduler scheduler, DataFileReader dataReader, 
			boolean containsTimestamps, ThreadGroup group, String id) {
		this(server, null, scheduler, dataReader, containsTimestamps, group, id, 1);
	}
	
	/**
	 * Constructor for sending events for both server and validator
	 * Event submission is controlled by the Scheduler passed as argument
	 * 
	 * @param server					A socket interface to the server/adapter 
	 * @param validator					A socket interface to a validator
	 * @param scheduler					Schedules event submission according to a given rate
	 * @param dataReader				Reads data sets from disk
	 * @param containsTimestamps		Indicates if the events in the data file are associated to timestamps
	 * @param validationSamplingRate	The percentage of events to forward to a validator
	 */
	public Sender(ClientSocketInterface server, ClientSocketInterface validator, Scheduler scheduler, 
			DataFileReader dataReader, boolean containsTimestamps, double validationSamplingRate) {
		this.server = server;	
		this.validator = validator;
		this.scheduler = scheduler;
		this.data = dataReader;		
		this.containsTimestamps = containsTimestamps;
		this.status = new Status(Step.STOPPED, 0);
		this.samplMod = (int) (1/validationSamplingRate);
	}
	
	/**
	 * Constructor for sending events only for server (no validation)
	 * Event submission is controlled by the Scheduler passed as argument
	 * 
	 * @param server				A socket interface to the server/adapter
	 * @param scheduler				Schedules event submission according to a given rate
	 * @param dataReader			Reads data sets from disk
	 * @param containsTimestamps	Indicates if the events in the data file are associated to timestamps
	 */
	public Sender(ClientSocketInterface server, Scheduler scheduler, DataFileReader dataReader, 
			boolean containsTimestamps) {
		this(server, null, scheduler, dataReader, containsTimestamps, 1);	
	}
	
	/**
	 * Constructor for sending events for both server and validator
	 * Event submission is controlled by timestamps in the datafile
	 * 
	 * @param server					A socket interface to the server/adapter
	 * @param validator					A socket interface to a validator
	 * @param dataReader				Reads data sets from disk
	 * @param timestampUnit				Time Unit of timestamps in data file 
	 * @param validationSamplingRate	The percentage of events to forward to a validator
	 */
	public Sender(ClientSocketInterface server, ClientSocketInterface validator, 
			DataFileReader dataReader, int timestampUnit, double validationSamplingRate) {
		this.scheduler = null;
		this.server = server;
		this.validator = validator;
		this.data = dataReader;
		this.samplMod = (int) (1/validationSamplingRate);
		this.containsTimestamps = true;
		this.timestampUnit = timestampUnit;
		this.status = new Status(Step.STOPPED, 0);		
	}
	
	/**
	 * Constructor for sending events only for server (no validation)
	 * Event submission is controlled by timestamps in the datafile
	 * 
	 * @param server				A socket interface to the server/adapter
	 * @param dataReader			Reads data sets from disk
	 * @param timestampUnit				Time Unit of timestamps in data file
	 */
	public Sender(ClientSocketInterface server, DataFileReader dataReader, int timestampUnit) {
		this(server, null, dataReader, timestampUnit, 1);
	}
	
	/**
	 * Return the current status of this sender thread
	 * 
	 * @return
	 */
	public Status getStatus() {
		return status;
	}
	
	/**
	 * Return the number of events sent so far by this sender thread
	 * 
	 * @return
	 */
	public long getSentEventCount() {
		return sentEventCount;
	}
	
	/**
	 * Stops sending events until a call to 'resume' method be done
	 */
	public void pauseLoad() {		
		this.status.setStep(Step.PAUSED);
	}
	
	/**
	 * Restarts event submission, if previously paused
	 */
	public void resumeLoad() {
		synchronized (this) {
			this.status.setStep(Step.RUNNING);
			notifyAll();
		}
	}
	
	/**
	 * Stops permanently event submission
	 */
	public void stopLoad() {		
		synchronized (this) {
			this.status.setStep(Step.STOPPED);
			// Case it is paused
			notifyAll();
		}
		
	}
	
	/**
	 * Method used to increase or decrease the rate at which events are submitted
	 * 
	 * @param factor		The multiplication factor (e.g. 2 = 2x faster than original rate)
	 */
	public void setRateFactor(double factor){
		this.scheduler.setRateFactor(factor);
	}
	
	public void run() {
		if(this.scheduler != null)
			scheduledRun();
		else
			timestampedRun();
	}
	
	private void scheduledRun() {
		long elapsedTime, interTime;
		long expectedElapsedTime = 0;
		long cumulateInterTime = 0;		
		long sleepTime;	
		long pauseT0;
		long timeResolution = (long) 1E6; // nanosecond/milisecond
		long sendTimestamp;
		
		try {
			//Event e = data.getNextEvent();
			this.status.setStep(Step.RUNNING);
			String event = data.getNextStringEvent();
			
			long t0 = System.currentTimeMillis();
			while(event != null) {
				// Removes event's timestamp, if it exists
				if(containsTimestamps)
					event = event.substring(event.indexOf(Globals.CSV_SEPARATOR)+1);
				
				//Checks if driver was paused and waits if so
				synchronized (this) {
					pauseT0 = 0;
					while(this.status.getStep() == Step.PAUSED) {
						if(pauseT0 == 0)
							pauseT0 = System.currentTimeMillis();
						this.wait();
					}
					if(pauseT0 != 0)
						timeInPause+=(System.currentTimeMillis()-pauseT0);					
				}
				
				//Checks if driver was stopped
				if(this.status.getStep() == Step.STOPPED)
					return;
				
				try {					
					/* Sleeps for a interval defined by scheduler according to event rate.
					 * The effective sleep time is adjusted to compensate for possible delays; 
					 * Actual Sleep time = interarrival time minus the difference between the actual elapsed time and the expected elapsed time.
					 */					
					interTime = this.scheduler.getInterArrivalTime();
					cumulateInterTime += interTime;		
					
					if(cumulateInterTime >= timeResolution) {					
						elapsedTime = System.currentTimeMillis()-t0;
						sleepTime = cumulateInterTime/timeResolution-((elapsedTime)-timeInPause-(expectedElapsedTime/timeResolution));
						cumulateInterTime = cumulateInterTime%timeResolution;						
						if(sleepTime > 0) {							
							Thread.sleep(sleepTime);
						}
							
					}			
		
					expectedElapsedTime = expectedElapsedTime + interTime;
					
					// Send the event to server;	
					sendTimestamp = System.currentTimeMillis();
					this.server.sendStringEvent(event);	
					
					sentEventCount++;
					
					try {
						// Send the event to validator
						if(this.validator != null && sentEventCount%this.samplMod == 0) {						
							this.validator.sendStringEventWithTimestamp(event, sendTimestamp);	
						}
					} catch (Exception ioe1) {
						System.err.println("Cannot send event to validator(" + ioe1.getMessage() + ")");
					}
					
				}
				catch (Exception ioe2) {
					System.err.println("Cannot send event (" + ioe2.getMessage() + ")");
					if(this.status.getStep() == Step.RUNNING)
						this.status.setStep(Step.ERROR);
				}
					
				event = data.getNextStringEvent();
			}	
			
			this.status.setStep(Step.FINISHED);
		}	
		catch (IOException ioe) {
			System.err.println("Cannot read datafile (" + ioe.getMessage() + ")");
			this.status.setStep(Step.ERROR);
		}
		catch (Exception exc) {
			System.err.println("Unexpected exception. (" + exc.getClass() + "-" + exc.getMessage() + ")\n load submisson will abort.");
			this.status.setStep(Step.ERROR);
		}
	}
	
	private void timestampedRun() {
		long elapsedTime, interTime;
		long expectedElapsedTime = 0;		
		long sleepTime;	
		long pauseT0;		
		long currentTS, lastTS;
		long timeResolution;
		long sendTimestamp;
		
		if(this.timestampUnit == ExternalFileWorkloadPhase.MILLISECONDS)
			timeResolution = 1;
		else // seconds and date/time
			timeResolution = 1000;
			
		try {
			//Event e = data.getNextEvent();
			this.status.setStep(Step.RUNNING);
			String event = data.getNextStringEvent();			
			String timestamp;
			
			// initializes lastTS variable
			if(event != null) {
				timestamp = event.substring(0, event.indexOf(Globals.CSV_SEPARATOR));				
				if(timestampUnit == ExternalFileWorkloadPhase.DATE_TIME)
					lastTS = Globals.DATE_TIME_FORMAT.parse(timestamp).getTime();
				else
					lastTS = Long.parseLong(timestamp);
			}
			else
				return;
			
			
			long t0 = System.currentTimeMillis();
			while(event != null) {
				timestamp = event.substring(0, event.indexOf(Globals.CSV_SEPARATOR));
				// Removes event's timestamp				
				event = event.substring(event.indexOf(Globals.CSV_SEPARATOR)+1);				
				
				if(timestampUnit == ExternalFileWorkloadPhase.DATE_TIME)
					currentTS = Globals.DATE_TIME_FORMAT.parse(timestamp).getTime();
				else
					currentTS = Long.parseLong(timestamp);
				
				//Checks if driver was paused and waits if so
				synchronized (this) {
					pauseT0 = 0;
					while(this.status.getStep() == Step.PAUSED) {
						if(pauseT0 == 0)
							pauseT0 = System.currentTimeMillis();
						this.wait();
					}
					if(pauseT0 != 0)
						timeInPause+=(System.currentTimeMillis()-pauseT0);					
				}
				
				//Checks if driver was stopped
				if(this.status.getStep() == Step.STOPPED)
					return;
				
				try {					
					/* Sleeps for a interval between consecutive events in the data file.
					 * The effective sleep time is adjusted to compensate for possible delays; 
					 * Actual Sleep time = interarrival time minus the difference between the actual elapsed time and the expected elapsed time.
					 */					
					interTime = timeResolution*(currentTS-lastTS);
												
					elapsedTime = System.currentTimeMillis()-t0;
					sleepTime = interTime - (elapsedTime-timeInPause-expectedElapsedTime);
					if(sleepTime > 0) {							
						Thread.sleep(sleepTime);
					}
		
					expectedElapsedTime = expectedElapsedTime + interTime;
					lastTS = currentTS;
					
					// Send the event to server;
					sendTimestamp = System.currentTimeMillis();
					this.server.sendStringEvent(event);	
					
					sentEventCount++;
					
					try {
						// Send the event to validator
						if(this.validator != null && sentEventCount%this.samplMod == 0) {						
							this.validator.sendStringEventWithTimestamp(event, sendTimestamp);	
						}
					} catch (Exception ioe1) {
						System.err.println("Cannot send event to validator(" + ioe1.getMessage() + ")");
					}
					
				}
				catch (Exception ioe2) {
					System.err.println("Cannot send event (" + ioe2.getMessage() + ")");
					if(this.status.getStep() == Step.RUNNING)
						this.status.setStep(Step.ERROR);
				}
					
				event = data.getNextStringEvent();
			}	
			
			this.status.setStep(Step.FINISHED);
		}	
		catch (IOException ioe) {
			System.err.println("Cannot read datafile (" + ioe.getMessage() + ")");
			this.status.setStep(Step.ERROR);
		}
		catch (Exception exc) {
			System.err.println("Unexpected exception. (" + exc.getClass() + "-" + exc.getMessage() + ")\n load submisson will abort.");
			this.status.setStep(Step.ERROR);
		}
	}
	
}
