package sink;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.net.ServerSocket;
import java.net.UnknownHostException;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Date;

import javax.swing.BorderFactory;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.Timer;
import javax.swing.border.EtchedBorder;

import basic.Globals;
import basic.InvalidStateException;
import basic.Status;
import basic.Step;

import communication.ClientSocketInterface;
import communication.SocketWorkerThread;

import data.CSVWriter;

@SuppressWarnings("serial")
public class Sink extends JFrame implements SinkRemoteFunctions{
	//========================= GUI ===================================
	JLabel statusLabel, eventCountLabel; 	
	JPanel statusPanel;
	JTextArea infoArea;	
	//=================================================================
	
	private ServerSocket incomingEventsSocket; //Used to receive events from adapters
	private AdapterListener workerThreads[];
	private CSVWriter logger;
	
	private String alias;
	
	private int samplMod;
	
	private ClientSocketInterface validator;
	
	private Status status;
	
	public Sink() {
		this("Sink");
	}
	
	public Sink(String alias) {
		super(alias);		
		this.alias = alias;
		this.status = new Status();
			
		statusPanel =  new JPanel();
		statusPanel.setLayout(null);
        statusPanel.setPreferredSize(new Dimension(200,400));
        statusPanel.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED), "Status"));
        statusLabel = new JLabel();                       
        eventCountLabel = new JLabel("Received events: 0");
        
		statusPanel.add(statusLabel);		
		statusPanel.add(eventCountLabel);
		
		statusLabel.setBounds(10, 10, 150, 50);
		eventCountLabel.setBounds(20, 45, 150, 50);
        
		infoArea = new JTextArea(10, 30);   
        infoArea.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED), "Info"));
        JScrollPane infoScroll = new JScrollPane(infoArea);
        infoArea.setEditable(false);        
        
		this.getContentPane().setLayout(new BorderLayout());
		this.getContentPane().add(statusPanel, BorderLayout.LINE_START);
		this.getContentPane().add(infoScroll, BorderLayout.CENTER);
		
		this.setSize(700, 400);        
		this.setLocationRelativeTo(null); //screen center
		this.setResizable(false);
		this.setDefaultCloseOperation(JFrame.HIDE_ON_CLOSE);
		this.setVisible(true);
		
		refreshGUI();						
		
		int delay = 1000/Globals.DEFAULT_GUI_REFRESH_RATE;
		Timer guiRefresher = new Timer(delay, new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent e) {				
				refreshGUI();					
			}			
		});
		guiRefresher.start();
		
		try {
			System.out.println("Sink application started. Initializing remote interface...");
			showInfo("Sink application started. Initializing remote interface...");
			this.initializeRMI();
			System.out.println("Done! Waiting for remote commands...");
			showInfo("Done! Waiting for remote commands...");
		} catch (Exception e) {
			this.status.setStep(Step.ERROR);			
			System.err.println("ERROR: Could not initialize remote interface: " + e.getMessage());
			showInfo("ERROR: Could not initialize remote interface: " + e.getMessage());
			try {
				Thread.sleep(5000);
			} catch (InterruptedException e1) {
				e1.printStackTrace();
			}
						
			guiRefresher.stop();
			guiRefresher = null;
			this.dispose();
			
		} 
		
	}
	
	private void initializeRMI() throws RemoteException {
		SinkRemoteFunctions stub = (SinkRemoteFunctions) UnicastRemoteObject.exportObject(this, 0);
		Registry registry = LocateRegistry.getRegistry();
		registry.rebind(alias, stub);	
	}

	public void refreshGUI() {
		switch (this.status.getStep()) {
		case DISCONNECTED:
			this.statusLabel.setIcon(Globals.YELLOW_SIGN);
			break;
		case CONNECTED:
		case READY:
		case PAUSED:
		case STOPPED:		
			this.statusLabel.setIcon(Globals.BLUE_SIGN);
			break;		
		case LOADING:
		case RUNNING: 
			this.statusLabel.setIcon(Globals.GREEN_SIGN);
			break;
		case ERROR:
			this.statusLabel.setIcon(Globals.RED_SIGN);
			break;
		}		
		
		if(this.workerThreads != null) {
			long receivedCount = 0;
			for (AdapterListener worker : workerThreads) {
				receivedCount+= worker.getReceivedEventCount();
			}
			eventCountLabel.setText("Received events: " + receivedCount);
			
			if(this.status.getStep() == Step.READY && receivedCount > 0)
				this.status.setStep(Step.RUNNING);
		}			
		else
			eventCountLabel.setText("Received events: 0");
		
		this.statusLabel.setText(this.status.getStep().toString());		
		
		statusPanel.revalidate();
	}
	

	public boolean load(SinkConfig mon) throws RemoteException, InvalidStateException {						
		if(this.status.getStep() == Step.DISCONNECTED ||
		   this.status.getStep() == Step.ERROR ||
		   this.status.getStep() == Step.STOPPED) {
			// Socket communication
			try {
				unload();
				this.status.setStep(Step.CONNECTED);
				incomingEventsSocket = new ServerSocket(mon.getPort());				
				String outputStreams[] = mon.getOutputStreamList();
				workerThreads = new AdapterListener[outputStreams.length];		
				for (int i = 0; i < workerThreads.length; i++) {
					workerThreads[i] = new AdapterListener("Listener for " + outputStreams[i], incomingEventsSocket);
					workerThreads[i].start();	
				}	
			} catch (IOException ioe) {	
				showInfo("Could not open server socket (" + ioe.getMessage() + ").");
				System.err.println("Could not open server socket (" + ioe.getMessage() + ").");
				this.status.setStep(Step.ERROR);
				return false;
			}	
		
					
			// Logging
			String logPath = Globals.APP_PATH + "log";
			File f = new File(logPath);
			if(!f.exists())
				f.mkdirs();
			
			try {
				logger = new CSVWriter(logPath + File.separator + mon.getAlias() + ".log");
			} catch (IOException ioe2) {
				System.err.println("Could not open logger (" + ioe2.getMessage() + ").");
				showInfo("Could not open logger (" + ioe2.getMessage() + ").");
				this.status.setStep(Step.ERROR);
				return false;
			}
			
			// Validation
			if(mon.isValidationEnabled()) {
				this.samplMod = (int) (1/mon.getValidationSamplingRate());
				
				// Tries to connect to validator
				if(validator != null) {
					try {
						validator.disconnect();
						validator = null;
					} catch (IOException ioe3) {					
						System.err.println("Could not disconnect from validator. (" +ioe3.getMessage() +")");
						showInfo("Could not disconnect from validator. (" +ioe3.getMessage() +")");
						this.status.setStep(Step.ERROR);					
					}
				}
				try {				
					showInfo("Trying to establish connection to validator at " + mon.getValidatorAddress().getHostAddress() + ":" + mon.getValidatorPort() + "...");					
					this.validator = new ClientSocketInterface(mon.getValidatorAddress(), mon.getValidatorPort(), false, mon.getAlias());
					System.out.println("Done!");
					showInfo("Done!");
				} catch (UnknownHostException e) {						
					System.err.println("ERROR: Cannot connect to specified validator: Unknown host.");
					showInfo("ERROR: Cannot connect to specified validator: Unknown host.");
					this.status.setStep(Step.ERROR);
					return false;
				} catch (IOException e) {
					System.err.println("ERROR: Cannot connect to specified validator. (" + e.getMessage() + ").");
					showInfo("ERROR: Cannot connect to specified validator. (" + e.getMessage() + ").");
					this.status.setStep(Step.ERROR);
					return false;
				}
			}
			
			this.status.setStep(Step.READY);
			showInfo(alias+" has been successfully loaded.");
			return true;
		}
		else {
			throw new InvalidStateException("Could not load sink. Sink has already been loaded.");
		}			
	}
	

	public void unload() {
		if(workerThreads != null) {
			for (AdapterListener worker : workerThreads) {
				worker.close();
				worker = null;
			}
			workerThreads = null;	
		}
		
		if(this.incomingEventsSocket != null) {
			try {
				this.incomingEventsSocket.close();
			} catch (IOException e) {
				System.err.println("Exception while closing server socket (" + e.getMessage() + ").");
			}
			this.incomingEventsSocket = null;
		}
		
		if(this.logger != null) {
			try {
				this.logger.closeFile();
				this.logger = null;	


			} catch (IOException e) {
				System.err.println("Exception while closing log file + (" + e.getMessage() + ").");
			}
		}	
		
		if(this.validator != null) {
			try {
				validator.disconnect();
				validator = null;
			} catch (IOException e) {
				System.err.println("Exception while disconnecting from Validator.");
			}
		}
				
		this.status.setStep(Step.STOPPED);
	}

	
	public void showInfo(String msg) {
		Date now = new Date();		
		infoArea.append(Globals.TIME_FORMAT.format(now) + " - " + msg + "\n" );
		infoArea.setCaretPosition(infoArea.getDocument().getLength());
	}
	
	class AdapterListener extends SocketWorkerThread{
		private long receivedEvents = 0;
		
		public AdapterListener(String threadID, ServerSocket ss) {
			super(threadID, ss);
		}

		@Override
		public void processIncomingMessage(Object o) throws Exception {
			String event = (String)o;
			long timestamp = System.currentTimeMillis();
			logger.writeRecord(event, timestamp);
			receivedEvents++;
			// Send the event to validator
			if(validator != null) {
				try {					
					long causalID = Long.parseLong(event.substring(event.lastIndexOf(Globals.CSV_SEPARATOR)+1));;					
					if(causalID%samplMod == 0)
						validator.sendStringEventWithTimestamp((String) o, timestamp);	
				} catch (NumberFormatException nfe) {
					System.err.println("Could not forward event to validator; Causal ID is not present");
				} 
			}			
		}
		
		public long getReceivedEventCount() {			
			return receivedEvents;
		}
		
	}
	

	@Override
	public Status getStatus() throws RemoteException {
		return this.status;
	}

}
