package validation;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.UnknownHostException;
import java.text.DecimalFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.Set;
import java.util.Map.Entry;

import javax.swing.BorderFactory;
import javax.swing.DefaultCellEditor;
import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.SwingUtilities;
import javax.swing.Timer;
import javax.swing.border.EtchedBorder;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.swing.table.DefaultTableModel;

import org.jfree.data.time.TimeSeries;

import sink.SinkConfig;
import basic.Globals;

import communication.SocketWorkerThread;

import controller.ConfigurationParser;
import driver.DriverConfig;

/**
 * 
 * Sample Online Validation Tool: The FINCoS Performance Monitor. 
 * Receives events from Drivers and Sinks, 
 * correlates them, and computes performance metrics
 * (response time and throughput of continuous queries)
 * 
 * @author Marcelo R.N. Mendes
 *
 */
@SuppressWarnings("serial")
public class OnlineValidationTool extends JFrame{	
	long estimated = 0;
	long total = 0;
	
	long windowSize = 10000;

	HashMap<InetAddress, LinkedHashMap<String, Long>> lookupTables; // one for each engine
	HashMap<String, OnlineValidator> validators; // one for each query/engine
	private ArrayList<ServerSocket> ssList; 
	ArrayList<Receiver> workers;

	private ArrayList<DriverConfig> drivers;
	private ArrayList<SinkConfig> sinks;
	
	
	private ConfigurationParser config = new ConfigurationParser();
	
	//========================= GUI ===================================	
	
	JPanel configPanel, componentsPanel, statsPanel, driversPanel, sinksPanel, graphsPanel;		
	
	JTabbedPane centerPanel;
	JTextArea infoArea;
	
	ArrayList<JLabel> driverLabelList, sinkLabelList;
	
	JScrollPane statsScroll, graphsScroll;
	HashMap <String,JTable> statsTables;
	HashMap <String,TimeSeries> throughputGraphSeries, rtGraphSeries, inputThroughputGraphSeries;
	DecimalFormat tableFormat = new DecimalFormat("0.00");
	Font statsTableFont = new Font("arial", Font.PLAIN, 14);
	
	JTextField configFilePathField;
	JFileChooser chooser;
	
	Timer guiRefresher;	

	private static final int GUI_REFRESH_RATE = 1; // per second
	private static final ImageIcon GREEN_SIGN = new ImageIcon("imgs/green.png");	
	private static final ImageIcon YELLOW_SIGN = new ImageIcon("imgs/yellow.png");
	
	//=================================================================
	
	
	public OnlineValidationTool() {
		super("FINCoS Performance Monitor");
		initGUI();
	}


	private void initGUI() {
		Container c = this.getContentPane();
		c.setLayout(new BorderLayout());

		// Configuration File Panel
		configPanel = new JPanel();
		configPanel.setBorder(BorderFactory.createTitledBorder("Test Configuration"));
		configPanel.setPreferredSize(new Dimension(200, 75));

		chooser = new JFileChooser(Globals.APP_PATH+"config");
		chooser.addChoosableFileFilter(new FileNameExtensionFilter("XML Configuration file", "xml"));
		chooser.setAcceptAllFileFilterUsed(false);
		chooser.setFileSelectionMode(JFileChooser.FILES_ONLY);

		JLabel configFilePathLabel = new JLabel("Path: ");		
		configFilePathField = new JTextField(40);
		configFilePathField.setToolTipText("Path of file containing test setup");
		JButton browseBtn = new JButton("...");
		browseBtn.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent e) {
				int action = chooser.showOpenDialog(null);
				if(action == JFileChooser.APPROVE_OPTION && chooser.getSelectedFile() != null) {
					configFilePathField.setText(chooser.getSelectedFile().getPath());
					try {
						if(guiRefresher != null)
							guiRefresher.stop();
						config.open(chooser.getSelectedFile().getPath());
						
						InetAddress myAddress = InetAddress.getLocalHost();
						
						// Retrieve Drivers List
						DriverConfig allDrivers[] = config.retrieveDriverList();
						driversPanel.removeAll();
						driversPanel.setLayout(new GridLayout(allDrivers.length+5, 1));
						drivers = new ArrayList<DriverConfig>(allDrivers.length);
						driverLabelList= new ArrayList<JLabel>(allDrivers.length);
						for (DriverConfig dr : allDrivers) {		
							if( dr.isValidationEnabled() && 
							    dr.getValidatorAddress().equals(myAddress)
							    || 
							    (dr.getServerAddress().getHostAddress().equals("127.0.0.1") && 
							     (dr.getAddress().getHostAddress().equals("127.0.0.1")|| 
							      dr.getAddress().getHostAddress().equals(myAddress.getHostAddress())
							     )
							    )	
							   ) 
							{
								addDriver(dr);
							}							
						}
						driversPanel.revalidate();
						driversPanel.repaint();
						
						// Retrieve Sinks List
						SinkConfig allSinks[] = config.retrieveSinkList();					
						sinksPanel.setLayout(new GridLayout(allSinks.length+5, 1));
						sinksPanel.removeAll();
						sinks = new ArrayList<SinkConfig>(allSinks.length);
						sinkLabelList = new ArrayList<JLabel>(allSinks.length);
						for (SinkConfig sink : allSinks) {
							if(sink.isValidationEnabled() && 
							   sink.getValidatorAddress().equals(myAddress)
							   || 
							   (sink.getServerAddress().getHostAddress().equals("127.0.0.1") && 
								(sink.getAddress().getHostAddress().equals("127.0.0.1") || 
								 sink.getAddress().getHostAddress().equals(myAddress.getHostAddress())
								)
							   )		
							  ) 
							{
								addSink(sink);	
							}
						}
						sinksPanel.revalidate();
						sinksPanel.repaint();
						
						showInfo("Test setup loaded successfully");																	

					} catch (Exception exc) {			
						JOptionPane.showMessageDialog(null, "Could not open configuration file. File may be corrupted. ", "Error", JOptionPane.ERROR_MESSAGE);						
						return;
					} 
					
					if(drivers == null || drivers.isEmpty() ||
					   sinks == null || sinks.isEmpty()) {
						JOptionPane.showMessageDialog(null, "No component is configured to send events to this validation tool instance");
						return;
					}
						
					
					// Initializes communication					
					try {
						showInfo("Initializing communication...");
						openServerSockets();
						initializeWorkerThreads();
						showInfo("Done! Waiting for client connections.");
					} catch (IOException exc) {
						showInfo("Error: " + exc.getClass() + " message: " + exc.getMessage());						
					}
					
					// Initializes Lookup tables for input events
					initializeLookupTables();
					
					// Initializes Validation mapping stream -> Validator
					initializeValidatorsTable();
					
					// Load stats panel
					statsTables = new HashMap<String, JTable>();					
					loadStatsPanel();
					
					// Load graphs panel
					rtGraphSeries = new HashMap<String, TimeSeries>();
					throughputGraphSeries = new HashMap<String, TimeSeries>();
					inputThroughputGraphSeries = new HashMap<String, TimeSeries>();
					loadGraphsPanel();
					
					guiRefresher.start();
				}

			}

		});
		configPanel.add(configFilePathLabel);
		configPanel.add(configFilePathField);
		configPanel.add(browseBtn);

		// Components Panel
		componentsPanel = new JPanel();
		componentsPanel.setLayout(new BorderLayout());
		driversPanel = new JPanel();		
		JScrollPane driversScroll = new JScrollPane(driversPanel);
		driversPanel.setBorder(BorderFactory.createTitledBorder("Drivers"));
		driversPanel.setBackground(Color.WHITE);
		sinksPanel = new JPanel();
		JScrollPane sinksScroll = new JScrollPane(sinksPanel);
		sinksPanel.setBorder(BorderFactory.createTitledBorder("Sinks"));
		sinksPanel.setBackground(Color.WHITE);
		sinksPanel.setLayout(new BorderLayout());				
		JPanel componentsCenterPanel = new JPanel();
		componentsCenterPanel.setLayout(new GridLayout(1,2));
		componentsCenterPanel.add(driversScroll);
		componentsCenterPanel.add(sinksScroll); 
		infoArea = new JTextArea(8, 20);               
        infoArea.setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(EtchedBorder.LOWERED), "Info"));
        JScrollPane infoScroll = new JScrollPane(infoArea);
        infoArea.setEditable(false);		
		componentsPanel.add(componentsCenterPanel, BorderLayout.CENTER);
		componentsPanel.add(infoScroll, BorderLayout.PAGE_END);
		

		// Stats Panel
		statsPanel = new JPanel();
		statsScroll = new JScrollPane(statsPanel);

		
		// GraphsPanel
		graphsPanel = new JPanel();
		graphsScroll = new JScrollPane(graphsPanel);
		centerPanel = new JTabbedPane();
		centerPanel.addTab("Components", componentsPanel);
		centerPanel.addTab("Stats", statsScroll);	
		centerPanel.add("Graphs", graphsScroll);

		c.add(configPanel, BorderLayout.PAGE_START);
		c.add(centerPanel, BorderLayout.CENTER);


		this.setDefaultCloseOperation(EXIT_ON_CLOSE);

		this.addWindowListener(new WindowAdapter() {
			@Override
			public void windowClosing(WindowEvent e) {
				close();		
			}

		});

		int delay = 1000/GUI_REFRESH_RATE;
		guiRefresher = new Timer(delay, new ActionListener(){
			@Override
			public void actionPerformed(ActionEvent e) {				
				refreshGUI();					
			}		
		});
		guiRefresher.start();

		this.setSize(850, 700);
		this.setLocationRelativeTo(null);
		this.setVisible(true);		
	}

	private void loadStatsPanel() {
		statsPanel.removeAll();
		statsTables.clear();
		JPanel serverPanel; 
		JTable statsTable;
		InetAddress serverList []= getServerList();
		for (InetAddress serverAddress : serverList) {
			serverPanel = new JPanel(new BorderLayout());
	
			serverPanel.setBorder(BorderFactory.createTitledBorder("CEP Engine at " + serverAddress.getHostAddress()));
			statsTable = new JTable();
			DefaultTableModel model = new DefaultTableModel(new String [] {"Key", "Query", "Validator", "Avg Throughput", "Last Throughput", "Avg RT", "MinRT", "MaxRT", "Last RT"}, 0)	{
				@Override
				public boolean isCellEditable(int row, int column) {
						return column == 2; // Only the validator combo box cell is editable
				}
			};
			statsTable.setModel(model);			
			statsTable.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
			statsTable.setFont(statsTableFont);
			statsTable.setRowHeight(40);
			
			//hides column key (it is used only internally)
			statsTable.getColumnModel().getColumn(0).setPreferredWidth(0);
			statsTable.getColumnModel().getColumn(0).setMaxWidth(0);
			
			// Puts a combo box in the table so the user can choose the appropriate validator.
			// THIS PART HAS TO CHANGE WHEN NEW VALIDATORS ARE INCLUDED.
			JComboBox comboBox = new JComboBox();	
			comboBox.setName(serverAddress.getHostAddress());
			statsTables.put(comboBox.getName(), statsTable);			
			comboBox.addItem("Perfmon");				
			comboBox.addItemListener(new ItemListener() {

				@Override
				public void itemStateChanged(ItemEvent e) {
					JComboBox combo = (JComboBox) e.getSource();
					JTable statsTable = statsTables.get(combo.getName());
					int row = statsTable.getSelectedRow();		
					if (row != -1 && e.getStateChange()==ItemEvent.SELECTED &&  combo.hasFocus()) {
						String serverAndOutput = (String) statsTable.getValueAt(row, 0);
						String selected = (String) combo.getSelectedItem();
						
						if(selected == null || selected.equals("")) {
							validators.put(serverAndOutput, null);
						}
						else if(selected.equals("Perfmon") ) {
							SinkConfig sink = getSinkConfig(serverAndOutput);
							double samplingRate = sink.getValidationSamplingRate();							
							LinkedHashMap<String, Long> lookupTable = getLookupTable(sink.getServerAddress());						
							
							validators.put(serverAndOutput, new SampleOnlineValidator(lookupTable, samplingRate));							
						} 									
					}
				}
				
			});			

			statsTable.getColumnModel().getColumn(2).setCellEditor(new DefaultCellEditor(comboBox));
			statsTable.getColumnModel().getColumn(2).setPreferredWidth(150);			
			String server, outputName;
			for (String serverAndOutput: validators.keySet()) {
				server = serverAndOutput.split("/")[0];				
				outputName = serverAndOutput.split("/")[1];
				if(server.equals(serverAddress.getHostAddress())) {
					model.addRow(new Object[] {serverAndOutput, outputName,null, "0.0", "0.0", "0.0", "0.0", "0.0"});					
				}
				
			}
			
			JScrollPane statsScroll = new JScrollPane(); 
			statsScroll.setViewportView(statsTable);
			
			serverPanel.add(statsScroll, BorderLayout.CENTER);
			serverPanel.setPreferredSize(new Dimension(50, statsTable.getRowCount()*30));
			statsPanel.setLayout(new GridLayout(serverList.length, 1));
			statsPanel.add(serverPanel);			
		}	 
		statsPanel.revalidate();
		
	}
	
	private void loadGraphsPanel() {
		graphsPanel.removeAll();
		this.rtGraphSeries.clear();
		this.throughputGraphSeries.clear();
		this.inputThroughputGraphSeries.clear();
		
		String serverName, queryName;
					
		ArrayList<String> queryList = new ArrayList<String>();
		HashMap<String, PerfChartPanel> rtGraphs = new HashMap<String, PerfChartPanel>();
		HashMap<String, PerfChartPanel> throughputGraphs = new HashMap<String, PerfChartPanel>();
		for (String key : validators.keySet()) {
			serverName = key.split("/")[0];
			queryName = key.split("/")[1];
						
			if(!queryList.contains(queryName)) {
				queryList.add(queryName);
				PerfChartPanel rtGraph, throughputGraph;
				rtGraph = new PerfChartPanel(queryName, null, "Response Time (ms)", new DecimalFormat("0"));
				this.rtGraphSeries.put(key,rtGraph.addSeries(serverName));
				throughputGraph = new PerfChartPanel(queryName, null, "Throughput (events/sec)", new DecimalFormat("0.00"));
				this.throughputGraphSeries.put(key,throughputGraph.addSeries(serverName));
				this.inputThroughputGraphSeries.put(key,throughputGraph.addSeries(serverName+"(Input)"));
				
				rtGraphs.put(queryName, rtGraph);
				throughputGraphs.put(queryName, throughputGraph);
			}
			else {
				this.rtGraphSeries.put(key,rtGraphs.get(queryName).addSeries(serverName));
				this.throughputGraphSeries.put(key,throughputGraphs.get(queryName).addSeries(serverName));
				this.inputThroughputGraphSeries.put(key,throughputGraphs.get(queryName).addSeries(serverName+"(Input)"));
			}
		}		
		graphsPanel.setLayout(new GridLayout(queryList.size(), 2));
		
		for (String query : queryList) {
			graphsPanel.add(rtGraphs.get(query));
			graphsPanel.add(throughputGraphs.get(query));
		}
		
		graphsPanel.revalidate();
		graphsPanel.repaint();

	}

	private void addDriver(DriverConfig dr) {
		JLabel drLabel = new JLabel(dr.getAlias() + " at " + dr.getAddress().getHostAddress());
		drLabel.setIcon(YELLOW_SIGN);
		drLabel.setToolTipText("Disconnected");
		this.drivers.add(dr);
		this.driversPanel.add(drLabel);		
		this.driverLabelList.add(drLabel);
	}
	
	private void addSink(SinkConfig sink) {
		JLabel sinkLabel = new JLabel(sink.getAlias() + " at " + sink.getAddress().getHostAddress());
		sinkLabel.setIcon(YELLOW_SIGN);
		sinkLabel.setToolTipText("Disconnected");
		this.sinks.add(sink);
		this.sinksPanel.add(sinkLabel);
		this.sinkLabelList.add(sinkLabel);
	}

	public void showInfo(String msg) {
		Date now = new Date();
	
		infoArea.append(Globals.TIME_FORMAT.format(now) + " - " + msg + "\n" );
		infoArea.setCaretPosition(infoArea.getDocument().getLength());
	}
	
	private void updateStatus(String id, ImageIcon icon, String toolTip) {
		for (JLabel drLabel : driverLabelList) {
			if(drLabel.getText().equals(id)) {
				drLabel.setIcon(icon);
				drLabel.setToolTipText(toolTip);
				drLabel.revalidate();
				return;
			}
		}
		
		for (JLabel sinkLabel : sinkLabelList) {			
			if(sinkLabel.getText().equals(id)) {
				sinkLabel.setIcon(icon);
				sinkLabel.setToolTipText(toolTip);				
				sinkLabel.revalidate();
				return;
			}
		}
		
	}	
	
	private void refreshGUI() {
		if(this.validators != null) {
			OnlineValidator validator;	
			
			if(this.statsTables != null) {						
				JTable table;
				for (Entry<String, JTable> e : statsTables.entrySet()) {				
					table = e.getValue();
					
					String outputName;
					for (int i = 0; i < table.getRowCount(); i++) {
						outputName = (String) table.getValueAt(i, 0);
						validator = validators.get(outputName);		
						long minRT;
						if(validator != null) {
							validator.recomputeStats();
							table.setValueAt(tableFormat.format(validator.getAvgThroughput()), i, 3) ;	
							table.setValueAt(tableFormat.format(validator.getCurrentThroughput()), i, 4) ;
							table.setValueAt(tableFormat.format(validator.getAvgRT()), i, 5) ;
							minRT = validator.getMinRT();
							if(minRT != Long.MAX_VALUE)
								table.setValueAt(minRT, i, 6) ;
							table.setValueAt(validator.getMaxRT(), i, 7) ;
							table.setValueAt(validator.getCurrentRT(), i, 8) ;
						}
						else
						{
							table.setValueAt("0.0", i, 3) ;
							table.setValueAt("0.0", i, 4) ;
							table.setValueAt("0.0", i, 5) ;
							table.setValueAt("0.0", i, 6) ;
							table.setValueAt("0.0", i, 7) ;
							table.setValueAt("0.0", i, 8) ;
						}
					}			
				}	
			}
			
			if(this.rtGraphSeries != null) {
				for (Entry<String, TimeSeries> e : rtGraphSeries.entrySet()) {
					validator = this.validators.get(e.getKey());
					if(validator != null) {
						PerfChartPanel.updateChart(e.getValue(), 
													System.currentTimeMillis(), 
													validator.getCurrentRT());		
					}
								
				}
			}
			
			if(this.throughputGraphSeries != null ) {
				for (Entry<String, TimeSeries> e : throughputGraphSeries.entrySet()) {
					validator = this.validators.get(e.getKey());
					if(validator != null) {
						PerfChartPanel.updateChart(e.getValue(), 
								System.currentTimeMillis(), 
								validator.getCurrentThroughput());	
					}
									
				}
			}
			
			if(this.inputThroughputGraphSeries != null ) {
				for (Entry<String, TimeSeries> e : inputThroughputGraphSeries.entrySet()) {
					validator = this.validators.get(e.getKey());
					if(validator != null) {
						PerfChartPanel.updateChart(e.getValue(), 
								System.currentTimeMillis(), 
								((SampleOnlineValidator)validator).getCurrentInputThroughput());	
					}
									
				}
			}
		}
				
	}	
	
	
	private void openServerSockets() throws IOException {
		close();
		
		Set<Integer> portList = new HashSet<Integer>(); 
		
		if(this.drivers != null) {
			for (DriverConfig dr : this.drivers) {
				portList.add(dr.getValidatorPort());
			}
		}
		
		if(this.sinks != null){
			for (SinkConfig sink : this.sinks) {
				portList.add(sink.getValidatorPort());
			}
		}
		
		this.ssList = new ArrayList<ServerSocket>(portList.size());
		for (Integer port : portList) {		
			try {
				showInfo("   Opening server socket at port " + port + "...");
				ServerSocket ss = new ServerSocket(port);
				ssList.add(ss);
				showInfo("   OK!");
			} catch (IOException e) {
				showInfo("Could not open server socket at " + port + "(" + e.getMessage() + ")");
			}
		
		}
	}
	
	private void close() {
		try {
			if(workers != null) {
				for (Receiver worker : workers) {
					if(worker != null) {
						worker.close();
						worker = null;
					}						
				}
				workers = null;
			}
			
			// Close server sockets
			if(ssList != null )
			{
				for (ServerSocket ss : ssList) {
					ss.close();	
				}
			}																
		} catch (IOException e1) {
			System.err.println("Could not close socket (" + e1.getMessage() + ")");
		}		
	}
	
	private void initializeWorkerThreads() {
		workers = new ArrayList<Receiver>(drivers.size()+sinks.size());
		Receiver r;		
		
		if(this.ssList != null) {
			if(this.drivers != null) {					
				for (DriverConfig dr : this.drivers) {
					for (ServerSocket ss : this.ssList) {
						if(dr.getValidatorPort() == ss.getLocalPort()) {
							r = new Receiver("ANONYMOUS", ss);
							workers.add(r);
							r.start();
							break;
						}
					}
					
				}
			}
			
			if(this.sinks != null){
				for (SinkConfig sink : this.sinks) {
					for (ServerSocket ss : this.ssList) {
						if(sink.getValidatorPort() == ss.getLocalPort()) {
							r = new Receiver("ANONYMOUS", ss);
							workers.add(r);
							r.start();
							break;
						}
					}
				}
			}	
		}
	
	}
	

	private SinkConfig getSinkConfig(String serverAndOutput) {				
		String server = serverAndOutput.split("/")[0];				
		String outputName = serverAndOutput.split("/")[1];
		
		for (SinkConfig sink : this.sinks) {
			if(server.equals(sink.getServerAddress().getHostAddress())) {
				for (String output : sink.getOutputStreamList()) {
					if(outputName.equals(output))
						return sink;
				} 
			}
					
		}
		
		return null;

	}
	
	private void initializeValidatorsTable() {
		validators = new HashMap<String, OnlineValidator>();
		
		String key;
		
		if(this.sinks != null){
			for (SinkConfig sink : this.sinks) {
				for (String outputStream : sink.getOutputStreamList()) {
					key = sink.getServerAddress().getHostAddress()+"/"+outputStream;
					validators.put(key, null);			
				}
						
			}
		}		
		
	}
	
	
	/**
	 * Creates and allocates one lookup table for each CEP engine.
	 * Lookup table contains input events used to compute response time. 
	 * 
	 */
	private void initializeLookupTables() {
		InetAddress[] serverList = getServerList();
		
		lookupTables = new HashMap<InetAddress, LinkedHashMap<String,Long>>(serverList.length);
		
		for (InetAddress serverAddress : serverList) {
			lookupTables.put(serverAddress, new LinkedHashMap<String, Long>());
		}
	}
	
	/**
	 * Retrieves the list of all CEP engines (Servers) used in the test from the drivers and sinks lists
	 * 
	 * @return	An array with the server list
	 */
	private InetAddress[] getServerList() {
		InetAddress ret[];
		Set<InetAddress> list = new HashSet<InetAddress>();
		
		if(this.drivers != null) {					
			for (DriverConfig dr : this.drivers) {
				list.add(dr.getServerAddress());
			}
		}
		
		if(this.sinks != null){
			for (SinkConfig sink : this.sinks) {
				list.add(sink.getServerAddress());
			}
		}
		
		ret = new InetAddress[list.size()];
		list.toArray(ret);
		
		return ret;
	}
	
	/**
	 * Gives the configuration of the component that just connected to the specified port from the specified address 
	 * 
	 * @param componentAddress		The address of the remote component
	 * @param localPort				The port on local machine on which the component connected 
	 * @return						The configuration of the component (either DriverConfig or SinkConfig)	
	 */
	private Object getComponentConfig(InetAddress componentAddress, int localPort) {		
		if(this.drivers != null) {					
			for (DriverConfig dr : this.drivers) {
				if(dr.getAddress().equals(componentAddress) && dr.getValidatorPort() == localPort)
					return dr;
			}

		}

		if(this.sinks != null){
			for (SinkConfig sink : this.sinks) {
				if(sink.getAddress().equals(componentAddress) && sink.getValidatorPort() == localPort)							
					return sink;
			}
		}

		return null;
	}
	
	/**
	 * Gives the appropriate lookup table for the component that just connected
	 * 
	 * @param serverAddress			The address of the server (CEP engine)	 
	 * @return						A pointer to the appropriate lookup table given the component's server (CEP engine) 
	 */
	private LinkedHashMap<String, Long> getLookupTable(InetAddress serverAddress) {
		return this.lookupTables.get(serverAddress);
	}
	

	class Receiver extends SocketWorkerThread{
		static final int INPUT = 0;
		static final int OUTPUT = 1;
		
		private int function;
		private InetAddress serverAddress;
		LinkedHashMap<String, Long> inputLookupTable;
		
		public Receiver(String threadID, ServerSocket ss)  {
			super(threadID, ss);			
		}
		
		public void setFunction(int function){
			if(function == INPUT || function == OUTPUT)
				this.function = function;
			else
				this.function = -1;
		}
		
		protected void setInputLookupTable(LinkedHashMap<String, Long> inputEventsTimestamps) {
			this.inputLookupTable = inputEventsTimestamps;
		}
		
		/**
		 * Stores input events for posterior computing of response times
		 */
		private void storeInputEvent(String record) {
			if(record == null)
				return;
			
			long timestamp= Long.parseLong(record.substring(0, 13));
			String ID = record.split(Globals.CSV_SEPARATOR)[2];				 			
			
			synchronized (inputLookupTable) {				
				if(inputLookupTable.size() > windowSize) {
					Iterator<Entry<String, Long>> i = inputLookupTable.entrySet().iterator();
					i.next();
					i.remove();									
				}
				
				inputLookupTable.put(ID, timestamp);
			}
		}
		
		/**
		 * Forwards output event for an appropriate Validator
		 * 
		 * @param event
		 */
		public void forwardToValidator(String event) {						
			String type = event.substring(event.indexOf(":")+1, event.indexOf(Globals.CSV_SEPARATOR, event.indexOf(Globals.CSV_SEPARATOR)+1));			
			String address = this.serverAddress.getHostAddress();
			String key;
			
			key = address + "/" + type;
					
			OnlineValidator myValidator = validators.get(key);
			
			if(myValidator != null) {				
				myValidator.storeOnBuffer(event);
			}
			else {
				System.out.println("Could not find appropriate validator for " + this.serverAddress.getHostAddress() + "/" + type);
			}			
		}
		

		@Override
		public void processIncomingMessage(Object o) throws Exception {			
			if(this.function == INPUT) {
				storeInputEvent((String) o);
			}
			else {
				forwardToValidator((String) o);
			}
			
		}
		
		@Override
		protected void listenToConnections() {				
			if(this.keepListening) {
				disconnect();
				try {						
					System.out.println("Worker thread " + this.getName() + " ready. Waiting for client connections...");
					updateStatus(this.getName(), YELLOW_SIGN, "Disconnected");	
					this.clientSocket = serverSocket.accept();
					Object config = getComponentConfig(clientSocket.getInetAddress(), serverSocket.getLocalPort());					
					
					if(config != null) {						
						if(config instanceof DriverConfig) {
							DriverConfig dr = (DriverConfig) config;
							this.setName(dr.getAlias()+ " at " + dr.getAddress().getHostAddress());
							this.setFunction(INPUT);		
							// Sets the lookup table for input events
							this.setInputLookupTable(getLookupTable(dr.getServerAddress()));							
						}
						else if (config instanceof SinkConfig) {
							SinkConfig sink = (SinkConfig) config;
							this.setName(sink.getAlias()+ " at " + sink.getAddress().getHostAddress());
							this.setFunction(OUTPUT);
							this.serverAddress = sink.getServerAddress();
							// Sets the lookup table for input events
							this.setInputLookupTable(getLookupTable(this.serverAddress));
							System.out.println("");
							// Clears eventual existing validators
							for (String outputName : sink.getOutputStreamList()) {
								String serverAndOutput = this.serverAddress+"/"+outputName;
								OnlineValidator validator = validators.get(serverAndOutput);
								if(validator != null)
									validator.clear();
								
							}
						}						
						this.input = new ObjectInputStream(clientSocket.getInputStream());				
						System.out.println("Worker thread " + this.getName() + " accepted client connection.");
						showInfo(this.getName() + " is now connected.");
						updateStatus(this.getName(), GREEN_SIGN, "Connected");
						
				
					}
					else {
						System.out.println("Could not identify component. Connection will now be reset.");
						listenToConnections();
					}
					
				} catch (IOException ioe) {
					System.err.println("Could not accept client connections. (" + ioe.getMessage() + ")");				
				}
								
			}
				
		}
		
	}
	
	public static void main(String[] args) throws UnknownHostException {
		SwingUtilities.invokeLater(new Runnable() {
			@Override
			public void run() {
				new OnlineValidationTool();
			}
			
		});			
	}

}
