package validation;

import java.util.ArrayList;
import java.util.LinkedHashMap;

/**
 * Abstract class representing a runtime Validator for maintaining
 * statistics about received events and performance counters. 
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public abstract class OnlineValidator {	
	long receivedCount=0;
	long processedCount=0;
	long elapsedTime=0;
	long totalRT=0;
	long maxRT=0;
	long minRT=Long.MAX_VALUE;
	double avgRT=0;
	long currentRT=0;
	double avgThroughput=0;
	double currentThroughput=0;
	LinkedHashMap<String, Long> lookupTable;
	
	ArrayList<String> outputBuffer; // Temporary storage for output events
	int outputBufferMAXSize; // for keeping output events
	static final int DEFAULT_OUTPUT_BUFFER_SIZE = 10000;
	
	public synchronized double getAvgThroughput() {
		return avgThroughput;
	}
	
	public double getCurrentThroughput() {
		return currentThroughput;
	}
	
	public synchronized double getAvgRT() {
		return avgRT;
	}
	public synchronized long getCurrentRT() {
		return currentRT;
	}
	
	public synchronized long getMaxRT() {
		return maxRT;
	}
	
	public synchronized long getMinRT() {
		return minRT;
	}
	
	public synchronized void clear() {
		receivedCount=0;
		processedCount=0;
		elapsedTime=0;
		totalRT=0;
		maxRT=0;
		minRT=Long.MAX_VALUE;
		avgRT=0;
		currentRT=0;
		avgThroughput=0;
		
		synchronized (outputBuffer) {
			outputBuffer.clear();
		}
	}
	
	public boolean storeOnBuffer(String event) {
		synchronized (outputBuffer) {
			if(this.outputBuffer.size() < this.outputBufferMAXSize) {
				return outputBuffer.add(event);			 
			}
			else
				return false;
		}
				
	}
	
	/**
	 * Recomputes the stats (throughput, RT) with output events stored in temporary buffer
	 * 
	 */
	public abstract void recomputeStats();	
}
