package validation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedList;

import basic.Globals;

/**
 * A sample Validator. Computes performance metrics in runtime 
 * from streams of events received from Drivers and Sinks. 
 * 
 * @author Marcelo R.N. Mendes
 *
 */
public class SampleOnlineValidator extends OnlineValidator{

	private int modFactor;	
	
	private LinkedList<Long> outputWindow; // Stores the timestamps of the last elements received from output streams for computing throughput	
	private int outputWindowSize; // for throughput computation
	private static final int DEFAULT_OUTPUT_WINDOW_SIZE = 100;
	
	long firstOutputEventTS, firstOutputWindowEventTS;
	long firstInputEventTS, firstInputWindowEventTS;
		
	
	private LinkedList<Long> inputWindow; // Stores the timestamps of the last elements received from output streams for computing throughput	

	private double currentInputThroughput;
	
	/**
	 * 
	 * @param lookupTable	A pointer to a memory structure containing timestamps of input events (used to compute response time)
	 * @param samplingRate	The sampling rate used by both Drivers and Monitors
	 */
	public SampleOnlineValidator(LinkedHashMap<String, Long> lookupTable, double samplingRate) {
		this(lookupTable, samplingRate, DEFAULT_OUTPUT_WINDOW_SIZE);
	}
	
	/**
	 * 
	 * @param lookupTable		A pointer to a memory structure containing timestamps of input events (used to compute response time)
	 * @param samplingRate		The sampling rate used by both Drivers and Monitors
	 * @param outputWindowSize	The number of output events to be kept when computing throughput 
	 */
	public SampleOnlineValidator(LinkedHashMap<String, Long> lookupTable, double samplingRate, int outputWindowSize) {
		this.modFactor = (int) (1/samplingRate);
		this.lookupTable = lookupTable;
		this.outputWindowSize = outputWindowSize;
		this.outputWindow = new LinkedList<Long>();
		this.outputBufferMAXSize = DEFAULT_OUTPUT_BUFFER_SIZE;
		this.outputBuffer = new ArrayList<String>(outputBufferMAXSize);
		
		this.inputWindow = new LinkedList<Long>();
	}
	
	@Override
	public void recomputeStats() {
		synchronized (outputBuffer) {
			Iterator<String> i = outputBuffer.iterator(); 
		
			while (i.hasNext()) {				
				if (recomputeStats(i.next()))
						i.remove();
			}			
		}
		
	}
	
	
	public boolean recomputeStats(String event) {
		receivedCount++;
				
		long inputID = Long.parseLong(event.substring(event.lastIndexOf(Globals.CSV_SEPARATOR)+1));
		Long inputTimestamp;
		synchronized (lookupTable) {
			inputTimestamp =  lookupTable.get(""+inputID);	
		}
		
		
		if(inputTimestamp != null) {			
			processedCount++;
			long outputTimestamp = Long.parseLong(event.substring(0, 13));
			this.outputWindow.offer(outputTimestamp);
			
			this.currentRT = outputTimestamp-inputTimestamp;
			if(currentRT >= 0)
				this.totalRT += this.currentRT;
			else {
				processedCount--;
				currentRT = minRT;
				System.out.println("Negative RT: " + currentRT + " ; Input: " + inputID +","+ inputTimestamp + " ; Output: " +outputTimestamp);
				return true;
			}
				
			this.avgRT = this.totalRT/processedCount;
			this.maxRT = Math.max(currentRT, maxRT);
			this.minRT = Math.min(currentRT, minRT);
						
			if(firstOutputEventTS == 0)
				firstOutputEventTS = outputTimestamp;
			
			elapsedTime = outputTimestamp-firstOutputEventTS;
			
			if(this.outputWindow.size() > this.outputWindowSize )
				 firstOutputWindowEventTS = outputWindow.poll();
			else
				firstOutputWindowEventTS = outputWindow.peek();
			
			int receivedWindow = modFactor*(outputWindow.size());
			long elapsedTimeWindow = (outputTimestamp-firstOutputWindowEventTS);
			try {
				this.currentThroughput = 1000.0*receivedWindow/elapsedTimeWindow;	
				this.avgThroughput = 1000.0*receivedCount*modFactor/elapsedTime;
			} catch (ArithmeticException e) {
				currentThroughput = 0;
				avgThroughput = 0;				
			}
			
			
			this.inputWindow.offer(inputTimestamp);
			if(this.inputWindow.size() > DEFAULT_OUTPUT_WINDOW_SIZE )
				 firstInputWindowEventTS = inputWindow.poll();
			else
				firstInputWindowEventTS = inputWindow.peek();
			
			int sentWindow = modFactor*(inputWindow.size());
			long elapsedInputTimeWindow = (inputTimestamp-firstInputWindowEventTS);
			try {
				this.currentInputThroughput = 1000.0*sentWindow/elapsedInputTimeWindow;
			} catch (ArithmeticException e) {
				currentInputThroughput = 0;						
			}
			
			return true;
		}	
		else {
			return false;
		}

	}
	
	public double getCurrentInputThroughput() {
		return currentInputThroughput;
	}
	
	@Override
	public void clear() {	
		System.out.println("clearing stats");
		super.clear();
			synchronized (lookupTable) {
				this.lookupTable.clear();	
			}
			this.firstOutputWindowEventTS = 0;
			this.firstOutputEventTS = 0;
	}
	

}

